# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function

import logging
import warnings

import numpy as np
import numpy.ma as ma

from decimal import Decimal
from argos.repo.memoryrtis import MappingRti, SyntheticArrayRti
from argos.repo.registry import ICON_COLOR_MEMORY, ICON_COLOR_PANDAS
from argos.utils.moduleinfo import versionStrToTuple


logger = logging.getLogger(__name__)

# the size of the colormap test images
SIZE_X = 500
SIZE_Y = 500

def makeConcentricCircles():
    """ Creates concentric cricle pattern.

        :return: 2D numpy array
    """
    x = np.linspace(-10, 10, num=SIZE_X)
    y = np.linspace(-10, 10, num=SIZE_Y)
    xx, yy = np.meshgrid(x, y, sparse=False)
    z = np.sin(xx**2 + yy**2) / (xx**2 + yy**2)
    return z


def makeRamp():
    """ Create 'ramp' function from Peter Karpov's blog http://inversed.ru/Blog_2.htm

        Slightly modified version of the test function introduced by Peter Kovesi
        Good Colour Maps: How to Design Them. Peter Kovesi, arxiv.org, 2015.
        http://arxiv.org/abs/1509.03700

        Allows to visually assess perceptual uniformity by observing the distance at which the sine \
        pattern fades.

        :return: 2D numpy array
    """
    x = np.linspace(0, 1, num=SIZE_X)
    y = np.linspace(0, 1, num=SIZE_Y)
    xx, yy = np.meshgrid(x, y, sparse=False)
    #z = yy + xx**2  # demonstrates banding
    z = yy + (xx**2) * np.sin(64 * 2 * np.pi * yy) / 12
    # z = np.clip(z, 0.0, 1.0) # Fails in PyQtGraph 2D plot :-/
    return z


def makeArcTan2():
    """ Create atan2(x, y), which is goog for testing circular color maps.
        :return: 2D numpy array
    """
    x = np.linspace(-1, 1, num=SIZE_X)
    y = np.linspace(-1, 1, num=SIZE_Y)
    xx, yy = np.meshgrid(x, y, sparse=False)
    return np.arctan2(xx, yy)


def makeSpiral():
    """ Create 'spiral' function from Peter Karpov's blog http://inversed.ru/Blog_2.htm

        Smoothness test, has a near-uniform distribution of values.
        :return: 2D numpy array
    """
    x = np.linspace(-1, 1, num=SIZE_X)
    y = np.linspace(-1, 1, num=SIZE_Y)
    xx, yy = np.meshgrid(x, y, sparse=False)
    return np.arcsin(np.sin(2 * 2 * np.pi * (xx**2 + yy**2) + np.arctan2(xx, yy)))


def makeSineProduct():
    """ Create 'two sines products' function from Peter Karpov's blog http://inversed.ru/Blog_2.htm

        Smoothness test.
        :return: 2D numpy array
    """
    x = np.linspace(-np.pi, np.pi, num=SIZE_X)
    y = np.linspace(-np.pi, np.pi, num=SIZE_Y)
    xx, yy = np.meshgrid(x, y, sparse=False)
    return np.sin(xx) * np.sin(yy) + np.sin(3*xx) * np.sin(3*yy)


def makeUniformNoise():
    """ Uniform noise between 0 and 1
    """
    return np.random.uniform(0.0, 1.0, size=(400, 300))


def createArgosTestData():
    """ Makes various test data sets for debugging
    """
    myDict = {}
    myDict['name'] = 'Pac Man'
    myDict['age'] = 34
    myDict['ghosts'] = ['Inky', 'Blinky', 'Pinky', 'Clyde']
    myDict['none'] = None
    myDict['numbers'] = {'int': 5, 'float': -6.6, 'large float': 7e77,
                         '-inf': np.NINF, 'nan': np.nan, 'complex': 8-9j, 'decimal': Decimal(4.444)}

    myDict['random'] = np.random.randn(8, 5)
    array = np.arange(240, dtype=np.float64).reshape(8, 30).transpose()
    #array[10, 4] = 7e77
    myDict['array'] = array

    myDict['ones'] = np.ones_like(array)

    x = np.arange(60)
    parabola = 0.02 * np.square(x)
    myDict['parabola'] = parabola

    myDict['array_1K'] = np.random.randn(1000)
    myDict['array_10K'] = np.random.randn(10 * 1000)
    myDict['array_100K'] = np.random.randn(100 * 1000)
    myDict['array_200K'] = np.random.randn(200 * 1000)
    myDict['array_500K'] = np.random.randn(500 * 1000)
    myDict['array_1000K'] = np.random.randn(1000 * 1000)

    #masked_array = ma.arange(2400, dtype=np.float16).reshape(60, 40)
    masked_array = np.transpose(ma.ones(shape=(40, 60)) * parabola)
    masked_array[:, 0:7] = ma.masked
    masked_array[15:45, 10:17] = ma.masked
    myDict['array_masked'] = masked_array

    nan_array = np.arange(2400, dtype=np.float16).reshape(60, 40)
    nan_array[:, 0:7] = np.nan
    nan_array[15:45, 10:17] = np.nan
    myDict['array_with_nans'] = nan_array

    inf_array = np.arange(2400, dtype=np.float16).reshape(60, 40)
    inf_array[:, 0:7] = np.inf
    inf_array[15:45, 10:17] = np.NINF
    myDict['array_with_infs'] = inf_array

    myDict['empty_array1D'] = np.ones(shape=(0, ))
    myDict['empty_array2D'] = np.ones(shape=(0, 6))
    myDict['empty_array3D'] = np.ones(shape=(10, 0, 6))

    myDict['structured_arr1'] = np.array([(1,2.,'Hello'), (2,3.,"World")],
                                          dtype=[('foo', 'i4'),('bar', 'f4'), ('baz', 'S10')])

    myDict['structured_arr2'] = np.array([(1.5,2.5,(1.0,2.0)),(3.,4.,(4.,5.)),(1.,3.,(2.,6.))],
                                         dtype=[('x','f4'),('y',np.float32),('value','f4',(2,2))])
    myDict['structured_arr3'] = np.array([(1.5,2.5,[2.0, -1.0]),(3.,4.,[5., 88]),(1.,3.,[2., 0.01])],
                                         dtype=[('1st','f4'),('2nd',np.float32),('3rd','f4',(2,))])

    # A structured array with offsets and titles
    dt4 = np.dtype({'names': ['r','b'], 'formats': ['u1', 'u1'], 'offsets': [0, 2]})
                    #'titles': ['Red pixel', 'Blue pixel']})
    myDict['structured_arr4'] = np.array([(255, 11), (1, 50)], dtype=dt4)

    # Nested structured array
    dt5 = [('year', '<i4'), ('countries', [('c1', [('iso','a3'), ('my_value','<f4')]),
                                           ('c2', [('iso','a3'), ('your_value','<f8', (3,))])])]
    myDict['structured_arr5'] = np.array([(2009, (('USA', 10.0), ('CHN', [-16.9, 0, 1000]))),
                                          (2010, (('BRA', 11.2), ('ARG', [12.0, 17.2, np.nan])))],
                                         dtype=dt5)

    # The _structured_masked_arr2 array fails for the Text inspector. Start with '_' to skip test.
    myDict['_structured_masked_arr2'] = ma.MaskedArray(myDict['structured_arr2'], fill_value=-99999)
    myDict['_structured_masked_arr2'].mask[0][0] = True
    myDict['_structured_masked_arr2'].mask[0][2][0,1] = True # doesn't seem to work

    myDict['numpy string array']  = np.array(['Yackity', 'Smackity'])
    myDict['numpy unicode array'] = np.array(['Table', u'ταБЬℓσ'])
    myDict['byte array'] = bytearray(range(256))
    myDict['bytes'] = bytes(bytearray(range(0, 256)))
    myDict['string'] = "The quick brown fox jumps over the lazy dog!"

    mappingRti = MappingRti(myDict, nodeName="myDict", iconColor=ICON_COLOR_MEMORY)

    # If set, the data below is added directly to the MappingRti instead to the underlying
    # dictionary. However, when you open and close the RTI in the tree, this data is lost.
    # Therefore, we turn it off for now and perhaps solve this problem later.
    includeVolatileData = False

    if includeVolatileData:
        # New axis at the beginning and end.
        mappingRti.insertChild(SyntheticArrayRti(
            'newaxis heading 2D', fun=lambda: makeConcentricCircles()[np.newaxis, np.newaxis, ...]))
        mappingRti.insertChild(SyntheticArrayRti(
            'newaxis trailing 2D', fun=lambda: makeConcentricCircles()[..., np.newaxis, np.newaxis]))

        mappingRti.insertChild(SyntheticArrayRti(
            'newaxis heading 1D', fun=lambda: makeConcentricCircles()[50,:][np.newaxis, np.newaxis, :]))

        mappingRti.insertChild(SyntheticArrayRti(
            'newaxis trailing 1D', fun=lambda: makeConcentricCircles()[50,:][:, np.newaxis, np.newaxis]))

        # Synthetic images for testing color maps.
        colorMapRti = mappingRti.insertChild(MappingRti({}, nodeName="test color maps"))
        colorMapRti.insertChild(SyntheticArrayRti('concentric circles', fun=makeConcentricCircles))
        colorMapRti.insertChild(SyntheticArrayRti('ramp', fun=makeRamp))
        colorMapRti.insertChild(SyntheticArrayRti('arctan2', fun=makeArcTan2))
        colorMapRti.insertChild(SyntheticArrayRti('spiral', fun=makeSpiral))
        colorMapRti.insertChild(SyntheticArrayRti('sine product', fun=makeSineProduct))
        colorMapRti.insertChild(SyntheticArrayRti('uniform noise', fun=makeUniformNoise))

        addPandasTestData(mappingRti)

    return mappingRti


def addPandasTestData(rti):
    """ Add some Pandas child RTIs to the rti
    """
    try:
        import pandas as pd
        from argos.repo.rtiplugins.pandasio import (PandasSeriesRti, PandasDataFrameRti,
                                                    PandasPanelRti)
    except ImportError as ex:
        logger.warning("No pandas test data created: {}".format(ex))
        return

    versionInfo = versionStrToTuple(pd.__version__)

    pandsRti = rti.insertChild(MappingRti({}, nodeName="_pandas"))  # for now skip in tests

    s = pd.Series([1, 2, 3, -4, 5], index=list('abcde'), name='simple series')
    pandsRti.insertChild(PandasSeriesRti(s, s.name, iconColor=ICON_COLOR_PANDAS))


    df = pd.DataFrame({'A' : ['foo', 'bar', 'foo', 'bar', 'foo', 'bar', 'foo', 'foo'],
                       'B' : ['one', 'one', 'two', 'three', 'two', 'two', 'one', 'three'],
                       'C' : np.random.randn(8),
                       'D' : np.random.randn(8)})

    pandsRti.insertChild(PandasDataFrameRti(df, 'df', iconColor=ICON_COLOR_PANDAS))

    # Multi index
    arrays = [['bar', 'bar', 'baz', 'baz', 'foo', 'foo', 'qux', 'qux'],
              ['one', 'two', 'one', 'two', 'one', 'two', 'one', 'two']]
    tuples = list(zip(*arrays))
    index = pd.MultiIndex.from_tuples(tuples, names=['first', 'second'])
    s = pd.Series(np.random.randn(8), index=index)
    pandsRti.insertChild(PandasSeriesRti(s, "multi-index", iconColor=ICON_COLOR_PANDAS))



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Miscellaneous routines.
"""
import logging, sys
import pprint

from html import escape

from argos.external.six import unichr
from argos.utils.cls import is_a_string

logger = logging.getLogger(__name__)

class NotSpecified(object):
    """ Class for NOT_SPECIFIED constant.
        Is used so that a parameter can have a default value other than None.

        Evaluate to False when converted to boolean.
    """
    def __nonzero__(self):
        """ Always returns False. Called when to converting to bool in Python 2.
        """
        return False

    def __bool__(self):
        """ Always returns False. Called when to converting to bool in Python 3.
        """
        return False



NOT_SPECIFIED = NotSpecified()



def python_major_version():
    """ Returns 2 or 3 for Python 2.x or 3.x respectively
    """
    return sys.version_info[0]


def python2():
    """ Returns True if we are running python 2
    """
    major_version = sys.version_info[0]
    assert major_version == 2 or major_version == 3, "major_version = {!r}".format(major_version)
    return major_version == 2


def is_quoted(s):
    """ Returns True if the string begins and ends with quotes (single or double)

        :param s: a string
        :return: boolean
    """
    return (s.startswith("'") and s.endswith("'")) or (s.startswith('"') and s.endswith('"'))



def string_to_identifier(s, white_space_becomes='_'):
    """ Takes a string and makes it suitable for use as an identifier

        Translates to lower case
        Replaces white space by the white_space_becomes character (default=underscore).
        Removes and punctuation.
    """
    import re
    s = s.lower()
    s = re.sub(r"\s+", white_space_becomes, s) # replace whitespace with underscores
    s = re.sub(r"-", "_", s) # replace hyphens with underscores
    s = re.sub(r"[^A-Za-z0-9_]", "", s) # remove everything that's not a character, a digit or a _
    return s


if __name__ == "__main__":
    print (string_to_identifier("Pea\nsdf-43q45,.!@#%&@&@@24n  pijn  Kenter, hallo$"))



def replaceStringsInDict(obj, old, new):
    """ Recursively searches for a string in a dict and replaces a string by another
    """
    if isinstance(obj, dict):
        return {key: replaceStringsInDict(value, old, new) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [replaceStringsInDict(value, old, new) for value in obj]
    elif is_a_string(obj):
        return obj.replace(old, new)
    else:
        return obj


def remove_process_serial_number(arg_list):
    """ Creates a copy of a list (typically sys.argv) where the strings that
        start with '-psn_0_' are removed.

        These are the process serial number used by the OS-X open command
        to bring applications to the front. They clash with argparse.
        See: http://hintsforums.macworld.com/showthread.php?t=11978
    """
    return [arg for arg in arg_list if not arg.startswith("-psn_0_")]


def replace_eol_chars(attr):
    """ Replace end-of-line characters with unicode glyphs so that all table rows fit on one line.
    """
    return (attr.replace('\r\n', unichr(0x21B5))
            .replace('\n', unichr(0x21B5))
            .replace('\r', unichr(0x21B5)))


def pformat(obj, width) -> str:
    """ Pretty print format with Argos default parameter values.
    """
    return pprint.pformat(obj, width=width, depth=2, sort_dicts=False)


def wrapHtmlColor(html, color):
    """ Wraps HTML in a span with a certain color
    """
    return '<span style="color:{}; white-space:pre;">{}</span>'\
        .format(color, escape(html, quote=False))




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Debug inspector.
"""
import logging

from argos.info import DEBUGGING
from argos.config.groupcti import MainGroupCti, GroupCti
from argos.inspector.abstract import AbstractInspector
from argos.qt import Qt, QtGui, QtWidgets

logger = logging.getLogger(__name__)


class DebugInspector(AbstractInspector):
    """ Empty inspector, mainly for debugging purposes.
    """
    def __init__(self, collector, parent=None):

        super(DebugInspector, self).__init__(collector, parent=parent)

        self.label = QtWidgets.QLabel()
        self.label.setWordWrap(True)
        self.contentsLayout.addWidget(self.label)

        self._config = self._createConfig()


    @classmethod
    def axesNames(cls):
        """ The names of the axes that this inspector visualizes.
            See the parent class documentation for a more detailed explanation.
        """
        return tuple()


    def _createConfig(self):
        """ Creates a config tree item (CTI) hierarchy containing default children.
        """
        rootItem = MainGroupCti('debug inspector')

        if DEBUGGING:
            # Some test config items.
            import numpy as np
            from argos.config.untypedcti import UntypedCti
            from argos.config.stringcti import StringCti
            from argos.config.intcti import IntCti
            from argos.config.floatcti import FloatCti, SnFloatCti
            from argos.config.boolcti import BoolCti, BoolGroupCti
            from argos.config.choicecti import ChoiceCti
            from argos.config.qtctis import PenCti

            grpItem = GroupCti("group")
            rootItem.insertChild(grpItem)

            lcItem = UntypedCti('line color', 123)
            grpItem.insertChild(lcItem)

            disabledItem = rootItem.insertChild(StringCti('disabled', "Can't touch me"))
            disabledItem.enabled=False

            grpItem.insertChild(IntCti('line-1 color', 7, minValue = -5, stepSize=2,
                                       prefix="@", suffix="%", specialValueText="I'm special"))
            rootItem.insertChild(StringCti('letter', 'aa', maxLength = 1))
            grpItem.insertChild(FloatCti('width', 2, minValue =5, stepSize=0.45, decimals=3,
                                         prefix="@", suffix="%", specialValueText="so very special"))
            grpItem.insertChild(SnFloatCti('scientific', defaultData=-np.inf))

            gridItem = rootItem.insertChild(BoolGroupCti('grid', True))
            gridItem.insertChild(BoolCti('X-Axis', True))
            gridItem.insertChild(BoolCti('Y-Axis', False))

            rootItem.insertChild(ChoiceCti('hobbit', 2, editable=True,
                                           configValues=['Frodo', 'Sam', 'Pippin', 'Merry']))
            myPen = QtGui.QPen(QtGui.QColor('#1C8857'))
            myPen.setWidth(2)
            myPen.setStyle(Qt.DashDotDotLine)
            rootItem.insertChild(PenCti('line', False, resetTo=myPen))

        return rootItem


    def _drawContents(self, reason=None, initiator=None):
        """ Draws the table contents from the sliced array of the collected repo tree item.

            The reason and initiator parameters are ignored.
            See AbstractInspector.updateContents for their description.
        """
        logger.debug("DebugInspector._drawContents: {}".format(self))

        slicedArray = self.collector.getSlicedArray()
        if slicedArray is None:
            text = "<None>"
        else:
            text = ("data = {!r}, masked = {!r}, fill_value = {!r} (?= {}: {})"
                    .format(slicedArray.data, slicedArray.mask, slicedArray.fill_value,
                            slicedArray.data.item(),
                            slicedArray.data.item() == slicedArray.fill_value))

        logger.debug("_drawContents: {}".format(text))
        logger.debug("_drawContents: {!r}".format(slicedArray))

        if DEBUGGING:
            self.label.setText(text)


    def _clearContents(self):
        """ Override to clear the widget contents.

            Is called to empty the inspector when an exception occurs during _drawContents,
            to prevent further errors.
        """
        logger.debug("DebugInspector._clearContents called")




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Plugins that use the PyQtGraph plot library.

    The default plugins are created (on registry reset) by InspectorRegistry.getDefaultItems()
"""
import logging

logger = logging.getLogger(__name__)

try:
    import pyqtgraph as pg
except Exception as ex:
    logger.exception(ex)
    raise ImportError("PyQtGraph 0.10.0 or higher required")


logger.debug("Imported PyQtGraph: {}".format(pg.__version__))


def setPgConfigOptions(**kwargs):
    """ Sets the PyQtGraph config options and emits a log message
    """
    for key, value in kwargs.items():
        logger.debug("Setting PyQtGraph config option: {} = {}".format(key, value))

    pg.setConfigOptions(**kwargs)


# Sets some config options
setPgConfigOptions(exitCleanup=False, crashWarning=True,
                   antialias=False,     # Anti aliasing of lines having width > 1 may be slow (OS-X)
                   leftButtonPan=True,  # If False, left button drags a rubber band for zooming
                   foreground='k',      # Default foreground color for axes, labels, etc.
                   background='w')      # Default background for GraphicsWidget





# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Defines ToggleColumnMixIn class
"""
from __future__ import print_function

import logging

from argos.qt import QtCore, QtWidgets, Qt

logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901


class LabeledWidget(QtWidgets.QWidget):
    """ A places a label and a widget next to each other.

        The layout will be created. It can be accessed as self.layout
        The label is passed as parameter and can be accessed as self.label
        The widget is passed as parameter and is available as self.widget
    """

    def __init__(self, label, widget,
                 layoutSpacing = None,
                 layoutContentsMargins = (0, 0, 0, 0),
                 labelStretch=0, spacerStretch=0, widgetStretch=1,
                 parent=None):
        """ Constructor.

            :param widget: the 'wrapped' widget, which will be displayed on the right
            :param labelText: text given to the label
            :param parent: parent wigdet
        """
        super(LabeledWidget, self).__init__(parent=parent)

        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(*layoutContentsMargins)
        if layoutSpacing is not None:
            self.layout.setSpacing(layoutSpacing)
        self.label = label
        self.widget = widget

        self.layout.addWidget(self.label, stretch=labelStretch)
        if spacerStretch:
            self.layout.addStretch(stretch=spacerStretch)
        self.layout.addWidget(self.widget, stretch=widgetStretch)



def labelTextWidth(label):
    """ Returns the width of label text of the label in pixels.

        IMPORTANT: does not work when the labels are styled using style sheets.

        Unfortunately it is possible to retrieve the settings (e.g. padding) that were set by the
        style sheet without parsing the style sheet as text.
    """
    # The Qt source shows that fontMetrics().size calls fontMetrics().boundingRect with
    # the TextLongestVariant included in the flags. TextLongestVariant is an internal flag
    # which is used to force selecting the longest string in a multi-length string.
    # See: http://stackoverflow.com/a/8638114/625350
    fontMetrics = label.fontMetrics()
    #contentsWidth = label.fontMetrics().boundingRect(label.text()).width()
    contentsWidth = fontMetrics.size(label.alignment(), label.text()).width()

    # If indent is negative, or if no indent has been set, the label computes the effective indent
    # as follows: If frameWidth() is 0, the effective indent becomes 0. If frameWidth() is greater
    # than 0, the effective indent becomes half the width of the "x" character of the widget's
    # current font().
    # See http://doc.qt.io/qt-4.8/qlabel.html#indent-prop
    if label.indent() < 0 and label.frameWidth(): # no indent, but we do have a frame
        indent = fontMetrics.width('x') / 2 - label.margin()
        indent *= 2 # the indent seems to be added to the other side as well
    else:
        indent = label.indent()

    result = contentsWidth + indent + 2 * label.frameWidth() + 2 * label.margin()

    if 1:
        #print ("contentsMargins: {}".format(label.getContentsMargins()))
        #print ("contentsRect: {}".format(label.contentsRect()))
        #print ("frameRect: {}".format(label.frameRect()))
        print ("contentsWidth: {}".format(contentsWidth))
        print ("lineWidth: {}".format(label.lineWidth()))
        print ("midLineWidth: {}".format(label.midLineWidth()))
        print ("frameWidth: {}".format(label.frameWidth()))
        print ("margin: {}".format(label.margin()))
        print ("indent: {}".format(label.indent()))
        print ("actual indent: {}".format(indent))
        print ("RESULT: {}".format(result))
        print ()
    return result


def labelsMaxTextWidth(labels):
    """ Returns the maximum width of the labels.
        IMPORTANT: does not work when the labels are styled using style sheets.
    """
    return max([labelTextWidth(label) for label in labels])


def harmonizeLabelsTextWidth(labels, width=None):
    """ Sets the the maximum width of the labels
        If width is None, the maximum width is calculated using labelsMaxTextWidth()
    """
    if width is None:
        width = labelsMaxTextWidth(labels)

    for label in labels:
        #label.setFixedWidth(width)
        label.setMinimumWidth(width)


if __name__ == "__main__":

    def _setLabelProps(label):
        #label.setFrameStyle(QtWidgets.QFrame.Box | QtWidgets.QFrame.Raised)
        label.setFrameStyle(QtWidgets.QFrame.Box | QtWidgets.QFrame.Plain)
        label.setLineWidth(1)

        #label.setMidLineWidth(10)
        #label.setMargin(25)
        #label.setIndent(4)
        #self.label.setAlignment(QtCore.Qt.AlignCenter)
        #self.label.setMaximumWidth(labelMinimumWidth)


    class MyTableView(QtWidgets.QTableView):

        def __init__(self):
            super(MyTableView, self).__init__()

            model = QtGui.QStandardItemModel(3, 2)
            self.setModel(model)
            self.horizontalHeader().resizeSection(0, 200)
            self.horizontalHeader().resizeSection(1, 300)




    def main():
        import sys

        app = QtWidgets.QApplication(sys.argv)
        window = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(window)

        if 0:
            if 1:
                window.setStyleSheet("""
                    QLabel {
                        background-color: #FF9900;
                    }
                """)
            else:
                window.setStyleSheet("""
                    QLabel {
                        margin: 5px;
                        border: 0px solid blue;
                        background-color: #FF9900;
                        padding: 0px;
                    }
                """)

        label0 = QtWidgets.QLabel('my great line edit')
        label1 = QtWidgets.QLabel('edit')
        label2 = QtWidgets.QLabel('combo')

        all_labels = [label0, label1, label2]
        for lbl in all_labels:
            _setLabelProps(lbl)
        harmonizeLabelsTextWidth(all_labels)

        maxWidth = labelsMaxTextWidth([label0, label1, label2])
        print("\mmaxWidth: {}".format(maxWidth))

        tableView =QtWidgets.QTableView()
        layout.addWidget(tableView)
        model = QtGui.QStandardItemModel(3, 2)
        tableView.setModel(model)
        tableView.horizontalHeader().resizeSection(0, 200)
        tableView.horizontalHeader().resizeSection(1, 300)
        layoutSpacing = 0

        editor0 = QtWidgets.QSpinBox()
        editor0.setValue(5)
        editor0.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        lw0 = LabeledWidget(label0, editor0, layoutSpacing=layoutSpacing)
        model.setData(model.index(0, 0), "A small")
        tableView.setIndexWidget(model.index(0, 1), lw0)

        editor1 = QtWidgets.QSpinBox()
        editor1.setValue(7)
        editor1.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.MinimumExpanding)
        lw1 = LabeledWidget(label1, editor1, layoutSpacing=layoutSpacing)
        model.setData(model.index(1, 0), "A SMALL seasoned curly")
        tableView.setIndexWidget(model.index(1, 1), lw1)

        comboBox = QtWidgets.QComboBox()
        comboBox.addItems(["Half diet coke", "Half regular coke", "Junior western bacon cheese"])
        lw2 = LabeledWidget(label2, comboBox, layoutSpacing=layoutSpacing)
        model.setData(model.index(2, 0), "What else?")
        tableView.setIndexWidget(model.index(2, 1), lw2)

        window.resize(550, 400)
        window.show()
        window.raise_()
        sys.exit(app.exec_())

    #######
    main()






# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Collector Tree.
"""
from __future__ import print_function

import logging
import warnings

import numpy as np

from argos.info import DEBUGGING
from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSlot
from argos.qt.togglecolumn import ToggleColumnTreeView
from argos.widgets.constants import COLLECTOR_TREE_ICON_SIZE
from argos.utils.cls import check_class

logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901

class CollectorTree(ToggleColumnTreeView):
    """ Tree widget for collecting the selected data. Includes an internal tree model.

        NOTE: this class is not meant to be used directly but is 'private' to the Collector().
        That is, plugins should interact with the Collector class, not the CollectorTree()
    """

    def __init__(self, parent):
        """ Constructor
        """
        super(CollectorTree, self).__init__(parent=parent)

        self._comboLabels = None

        nCols = 1
        model = QtGui.QStandardItemModel(3, nCols)
        self.setModel(model)
        self.setTextElideMode(Qt.ElideMiddle) # ellipsis appear in the middle of the text

        self.setRootIsDecorated(False) # disable expand/collapse triangle
        self.setUniformRowHeights(True)

        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setAnimated(True)
        self.setAllColumnsShowFocus(True)
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.setIconSize(COLLECTOR_TREE_ICON_SIZE)

        treeHeader = self.header()
        treeHeader.setStretchLastSection(False)
        treeHeader.setSectionsMovable(False)

        treeHeader.resizeSection(0, 600) # For item path
        treeHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive) # don't set to stretch

        labels = [''] * model.columnCount()
        labels[0] = "Path"
        model.setHorizontalHeaderLabels(labels)

        #enabled = dict((name, False) for name in self.HEADERS)
        #checked = dict((name, True) for name in self.HEADERS)
        #self.addHeaderContextMenu(checked=checked, enabled=enabled, checkable={})


    def resizeColumnsFromContents(self, startCol=None):
        """ Resize columns depending on their contents.

            The width of the first column (showing the path) will not be changed
            The columns containing combo boxes will be set to the size hints of these combo boxes
            The remaining width (if any) is devided over the spin boxes.
        """
        logger.debug("resizeColumnsFromContents called.")
        numCols = self.model().columnCount()
        startCol = 0 if startCol is None else max(startCol, 0)

        # Set columns with comboboxes to their size hints
        row = 0
        header = self.header()
        for col in range(startCol, numCols):
            indexWidget = self.indexWidget(self.model().index(row, col))
            if indexWidget:
                if isinstance(indexWidget, QtWidgets.QComboBox):
                    header.resizeSection(col, indexWidget.sizeHint().width())

        # Collect size hints of spin boxes and indices of all other columns.
        indexSpin = []
        indexNonSpin = []
        spinBoxSizeHints = []
        spinBoxMaximums = []
        for col in range(0, numCols):
            indexWidget = self.indexWidget(self.model().index(row, col))
            if indexWidget and isinstance(indexWidget, (QtWidgets.QSpinBox, SpinSlider)):
                spinBoxSizeHints.append(indexWidget.spinbox.sizeHint().width())
                spinBoxMaximums.append(max(0, indexWidget.spinbox.maximum())) # prevent negatives
                indexSpin.append(col)
            else:
                indexNonSpin.append(col)

        if len(indexSpin) == 0:
            return

        headerWidth = self.header().width()
        spinBoxSizeHints = np.array(spinBoxSizeHints)
        spinBoxTotalSizeHints = np.sum(np.array(spinBoxSizeHints))
        colWidths = np.array([self.header().sectionSize(idx) for idx in range(numCols)])
        nonSpinBoxTotalWidth = np.sum(colWidths[indexNonSpin])
        remainingTotal = max(0, headerWidth - nonSpinBoxTotalWidth - spinBoxTotalSizeHints)

        with warnings.catch_warnings():
            # Ignore divide by zero warnings when all elements have the same value
            warnings.simplefilter("ignore")
            spinBoxWeights = np.maximum(0.5, np.log10(np.array(spinBoxMaximums)))
            normSpinBoxWeights = spinBoxWeights / np.sum(spinBoxWeights)
            extraWidthPerSpinBox = remainingTotal * normSpinBoxWeights
            newSpinBoxWidths = spinBoxSizeHints + extraWidthPerSpinBox

        if DEBUGGING:  # Only during debugging to reduce devops log file size.
            logger.debug("Dividing the remaining width over the spinboxes.")
            logger.debug("Header width               : {}".format(headerWidth))
            logger.debug("Column widths              : {}".format(colWidths))
            logger.debug("Width of non-spinboxes     : {}".format(nonSpinBoxTotalWidth))
            logger.debug("Total size hint spinboxes  : {}".format(spinBoxTotalSizeHints))
            logger.debug("Remaining width to divide  : {}".format(remainingTotal))
            logger.debug("Spinbox maximums           : {}".format(spinBoxMaximums))
            logger.debug("Normalized spinbox weights : {}".format(normSpinBoxWeights))
            logger.debug("Extra width per spin box   : {}".format(extraWidthPerSpinBox))
            logger.debug("New spinbox widths         : {}".format(newSpinBoxWidths))

        # Divide the remaining width over the spin boxes using the log(nrElements) as weights.
        # If the remaining total is less than zero, just set the widths to the size hints (a
        # horizontal scrollbar will appear).
        for idx, newWidth in zip(indexSpin, newSpinBoxWidths):
            header.resizeSection(idx, round(newWidth))



class CollectorSpinBox(QtWidgets.QSpinBox):
    """ A spinbox for use in the collector.
        Overrides the sizeHint so that is not truncated when large dimension names are selected
    """
    def __init__(self, *args, **kwargs):
        """ Constructor
        """
        super(CollectorSpinBox, self).__init__(*args, **kwargs)
        self._cachedSizeHint = None


    def sizeHint(self):
        """ Reimplemented from the C++ Qt source of QAbstractSpinBox.sizeHint, but without
            truncating to a maximum of 18 characters.
        """
        # The cache is invalid after the prefix, postfix and other properties
        # have been set. I disabled it because sizeHint isn't called that often.
        #if self._cachedSizeHint is not None:
        #    return self._cachedSizeHint

        orgSizeHint = super(CollectorSpinBox, self).sizeHint()

        self.ensurePolished()
        fm = QtGui.QFontMetrics(self.fontMetrics())

        # This was h = d.edit.sizeHint().height(), but that didn't work. In the end we set the
        # height to the height calculated from the parent.
        h = orgSizeHint.height()
        w = 0

        # QLatin1Char seems not to be implemented.
        # Using regular string literals and hope for the best
        s = self.prefix() + self.textFromValue(self.minimum()) + self.suffix() #+ ' '

        # We disabled truncating the string here!!
        #s = s[:18]
        w = max(w, fm.width(s))

        s = self.prefix() + self.textFromValue(self.maximum()) + self.suffix() #+ ' '

        # We disabled truncating the string here!!
        #s = s[:18]
        w = max(w, fm.width(s))
        if len(self.specialValueText()):
            s = self.specialValueText()
            w = max(w, fm.width(s))

        w += 2 # cursor blinking space

        w -= 15 # The spinboxes seemed to wide. Made a bit smaller by Pepijn.

        opt = QtWidgets.QStyleOptionSpinBox()
        self.initStyleOption(opt)
        hint = QtCore.QSize(w, h)
        extra = QtCore.QSize(35, 6)

        opt.rect.setSize(hint + extra)
        extra += hint - self.style().subControlRect(
            QtWidgets.QStyle.CC_SpinBox, opt, QtWidgets.QStyle.SC_SpinBoxEditField, self).size()

        # get closer to final result by repeating the calculation
        opt.rect.setSize(hint + extra)
        extra += hint - self.style().subControlRect(
            QtWidgets.QStyle.CC_SpinBox, opt, QtWidgets.QStyle.SC_SpinBoxEditField, self).size()
        hint += extra

        opt.rect = self.rect()
        result = (self.style().sizeFromContents(QtWidgets.QStyle.CT_SpinBox, opt, hint, self)
                  .expandedTo(QtWidgets.QApplication.globalStrut()))
        self._cachedSizeHint = result

        # Use the height ancestor's sizeHint
        result.setHeight(orgSizeHint.height())

        return result


class SpinSlider(QtWidgets.QWidget):
    """ A SpinBox and Slider widgets next to each other.

        The layout will be created. It can be accessed as self.layout
    """
    #sigValueChanged = QtSignal(int)

    def __init__(self,
                 spinBox,
                 slider = None,
                 layoutSpacing = None,
                 layoutContentsMargins = (0, 0, 0, 0),
                 parent=None):
        """ Constructor.

            The settings (min, max, enabled, etc) from the SpinBox will be used for the slider
            as well. That is, the spin box is the master.
        """
        super(SpinSlider, self).__init__(parent=parent)

        check_class(spinBox, QtWidgets.QSpinBox, allow_none=True)
        check_class(slider, QtWidgets.QSlider, allow_none=True)

        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(*layoutContentsMargins)
        if layoutSpacing is not None:
            self.layout.setSpacing(layoutSpacing)

        if spinBox is None:
            self.spinbox = QtWidgets.QSpinBox()
        else:
            self.spinbox = spinBox

        if slider is None:
            self.slider = QtWidgets.QSlider(Qt.Horizontal)
        else:
            self.slider = slider

        self.slider.setMinimum(self.spinbox.minimum())
        self.slider.setMaximum(self.spinbox.maximum())
        self.slider.setValue(self.spinbox.value())
        self.slider.setEnabled(self.spinbox.isEnabled())

        self.layout.addWidget(self.spinbox, stretch=0)
        self.layout.addWidget(self.slider, stretch=1)

        self.spinbox.valueChanged.connect(self.slider.setValue)
        self.slider.valueChanged.connect(self.spinbox.setValue)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

"""
Class for storing values and a mask. Masked arrays would have been a good solution but
unfortunately they are very buggy.

TODO: what are these issues again?
"""
from __future__ import division, print_function
import logging
import warnings

import numpy as np
import numpy.ma as ma

from argos.utils.cls import check_class, is_an_array, check_is_an_array, array_is_structured

logger = logging.getLogger(__name__)


class ConsistencyError(Exception):
    """ Raised when the mask of an ArrayWithMask object has an inconsistent shape."""
    pass

# TODO: look at Numpy’s dispatch mechanism to see if we reimplement
# https://numpy.org/doc/stable/user/basics.dispatch.html#basics-dispatch
class ArrayWithMask(object):
    """ Class for storing an arrays together with a mask.

        Used instead of the Numpy MaskedArray class, which has too many issues.
    """
    def __init__(self, data, mask, fill_value):
        """ Constructor

            :param data:
            :param mask: array with mask or single boolean for the complete mask
            :param fill_value:
        """
        check_is_an_array(data)
        check_class(mask, (np.ndarray, bool, np.bool_))

        # Init fields
        self._data = None
        self._mask = None
        self._fill_value= None

        # Use setters
        self.data = data
        self.mask = mask
        self.fill_value= fill_value


    @property
    def data(self):
        """ The array values. Will be a numpy array."""
        return self._data


    @data.setter
    def data(self, values):
        """ The array values. Must be a numpy array."""
        check_is_an_array(values)
        self._data = values


    @property
    def mask(self):
        """ The mask values. Will be an array or a boolean scalar."""
        return self._mask


    @mask.setter
    def mask(self, mask):
        """ The mask values. Must be an array or a boolean scalar."""
        check_class(mask, (np.ndarray, bool, np.bool_))
        if isinstance(mask, (bool, np.bool_)):
            self._mask = bool(mask)
        else:
            self._mask = mask


    @property
    def fill_value(self):
        """ The fill_value."""
        return self._fill_value


    @fill_value.setter
    def fill_value(self, fill_value):
        """ The fill_value."""
        self._fill_value = fill_value


    def checkIsConsistent(self):
        """ Raises a ConsistencyError if the mask has an incorrect shape.
        """
        if is_an_array(self.mask) and self.mask.shape != self.data.shape:
            raise ConsistencyError("Shape mismatch mask={}, data={}"
                                   .format(self.mask.shape, self.data.shape))


    @classmethod
    def createFromMaskedArray(cls, masked_arr):
        """ Creates an ArrayWithMak

            :param masked_arr: a numpy MaskedArray or numpy array
            :return: ArrayWithMask
        """
        if isinstance(masked_arr, ArrayWithMask):
            return masked_arr

        check_class(masked_arr, (np.ndarray, ma.MaskedArray))

        # A MaskedConstant (i.e. masked) is a special case of MaskedArray. It does not seem to have
        # a fill_value so we use None to use the default.
        # https://docs.scipy.org/doc/numpy/reference/maskedarray.baseclass.html#numpy.ma.masked
        fill_value = getattr(masked_arr, 'fill_value', None)

        return cls(masked_arr.data, masked_arr.mask, fill_value)


    def asMaskedArray(self):
        """ Creates converts to a masked array
        """
        return ma.masked_array(data=self.data, mask=self.mask, fill_value=self.fill_value)


    def maskAt(self, index):
        """ Returns the mask at the index.

            It the mask is a boolean it is returned since this boolean representes the mask for
            all array elements.
        """
        if isinstance(self.mask, bool):
            return self.mask
        else:
            return self.mask[index]


    def maskIndex(self):
        """ Returns a boolean index with True if the value is masked.

            Always has the same shape as the maksedArray.data, event if the mask is a single boolan.
        """
        if isinstance(self.mask, bool):
            return np.full(self.data.shape, self.mask, dtype=np.bool)
        else:
            return self.mask


    @property
    def shape(self):
        """ Convenience method, returns the shape of the data.
        """
        return self.data.shape


    @property
    def dtype(self):
        """ Convenience method, returns the dtype of the data.
        """
        return self.data.dtype


    def __getitem__(self, index):
        """ Called when using the awm with an index (e.g. rti[0]).
        """
        newMask = self.mask if isinstance(self.mask, bool) else self.maskIndex()[index]
        return ArrayWithMask(self.data[index], newMask, fill_value=self.fill_value)


    def transpose(self, *args, **kwargs):
        """ Transposes the array and mask separately

            :param awm: ArrayWithMask
            :return: copy/view with transposed
        """
        tdata = np.transpose(self.data, *args, **kwargs)
        tmask = np.transpose(self.mask, *args, **kwargs) if is_an_array(self.mask) else self.mask
        return ArrayWithMask(tdata, tmask, self.fill_value)


    def replaceMaskedValue(self, replacementValue):
        """ Replaces values where the mask is True with the replacement value.
        """
        self.data = replaceMaskedValue(self.data, self.mask, replacementValue)


    def replaceMaskedValueWithNan(self):
        """ Replaces values where the mask is True with the replacement value.

            Will change the data type to float if the data is an integer.
            If the data is not a float (or int) the function does nothing.
        """
        replaceMaskedValueWithFloat(self.data, self.mask, np.nan, copyOnReplace=False)


#############
# functions #
#############

def replaceMaskedValue(data, mask, replacementValue, copyOnReplace=True):
    """ Replaces values where the mask is True with the replacement value.

        :copyOnReplace makeCopy: If True (the default) it makes a copy if data is replaced.
    """
    if mask is False:
        result = data
    elif mask is True:
        result = np.copy(data) if copyOnReplace else data
        result[:] = replacementValue
    else:
        #logger.debug("############ count_nonzero: {}".format(np.count_nonzero(mask)))
        if copyOnReplace and np.any(mask):
            #logger.debug("Making copy")
            result = np.copy(data)
        else:
            result = data

        result[mask] = replacementValue

    return result


def replaceMaskedValueWithFloat(data, mask, replacementValue, copyOnReplace=True):
    """ Replaces values where the mask is True with the replacement value.

        Will change the data type to float if the data is an integer.
        If the data is not a float (or int) the function does nothing. Otherwise it will call
        replaceMaskedValue with the same parameters.

        :copyOnReplace makeCopy: If True (the default) it makes a copy if data is replaced.
    """
    kind = data.dtype.kind
    if kind == 'i' or kind == 'u': # signed/unsigned int
        data = data.astype(np.float, casting='safe')

    if data.dtype.kind != 'f':
        return # only replace for floats
    else:
        return replaceMaskedValue(data, mask, replacementValue, copyOnReplace=copyOnReplace)


def nanPercentileOfSubsampledArrayWithMask(arrayWithMask, percentiles, subsample, *args, **kwargs):
    """ Sub samples the array and then calls maskedNanPercentile on this.

        If subsample is False, no sub sampling is done and it just calls maskedNanPercentile
    """
    check_class(subsample, bool)
    check_class(arrayWithMask, ArrayWithMask)

    maskedArray = arrayWithMask.asMaskedArray()

    if subsample:
        maskedArray = _subsampleArray(maskedArray)

    return _maskedNanPercentile(maskedArray, percentiles, *args, **kwargs)



def _subsampleArray(array, targetNumElements=40000):
    """ Sub samples an array or masked array.

        It samples the array down equally along all dimensions so that is has less than 40000
        elements. The 40000 is 200 by 200 for a 2D array, which was taken from the histogram
        subsampling of PyQtGraph. It works well but other values may work as well.

        If the array is already smaller than 40000 elements, no subsampling will be done.
    """
    check_class(array, np.ndarray)
    # logger.debug("_subsampleArray: {}, shape={}, dtype={}"
    #              .format(type(array), array.shape, array.dtype))
    oldShape = array.shape
    numDims = len(oldShape)

    targetDimSize = targetNumElements ** (1/numDims)
    steps = [max(1, int(np.ceil(shp / targetDimSize))) for shp in oldShape]

    slices = tuple(slice(None, None, step) for step in steps)

    result = array[slices]
    # logger.debug("Subsampling array shapes {} -> {}".format(oldShape, result.shape))
    return result



def _maskedNanPercentile(maskedArray, percentiles, *args, **kwargs):
    """ Calculates np.nanpercentile on the non-masked values

        The *args and **kwargs are passed on to np.nanpercentile
    """
    check_class(maskedArray, ma.masked_array)

    #https://docs.scipy.org/doc/numpy/reference/maskedarray.generic.html#accessing-the-data
    awm = ArrayWithMask.createFromMaskedArray(maskedArray)

    maskIdx = awm.maskIndex()
    validData = awm.data[~maskIdx]

    if len(validData) >= 1:
        with warnings.catch_warnings():
            # Suppress warnings when stepData consists of only infinite data
            warnings.simplefilter("ignore")
            result = np.nanpercentile(validData, percentiles, *args, **kwargs)

    else:
        # If np.nanpercentile on an empty list only returns a single Nan. We correct this here.
        result = len(percentiles) * [np.nan]

    assert len(result) == len(percentiles), \
        "shape mismatch: {} != {}".format(len(result), len(percentiles))

    return result # as expected



#
# def replace_missing_values(array, missing_value, replacement_value):
#     """ Returns a copy of the array where the missing_values are replaced with replacement_value.
#
#         If missing_value is None, nothing is replaced, a copy of the array is returned..
#         The missing_value can be Nan or infinite, these are replaced.
#
#         If array is a masked array the masked value are replaced (masked_to_regular_array).
#     """
#     if isinstance(array, ma.MaskedArray):
#         logger.debug("^^^^^^^^^^^^^^^^^^^ mask: {}".format(array.mask))
#
#         return array
#         return masked_to_regular_array(array, replacement_value=replacement_value)
#     if missing_value is None:
#         array = np.copy(array)
#     elif np.isnan(missing_value):
#         array[np.isnan(array)] = replacement_value
#     else:
#         array[array == missing_value] = replacement_value
#
#     return array


def fillValuesToNan(masked_array):
    """ Replaces the fill_values of the masked array by NaNs

        If the array is None or it does not contain floating point values, it cannot contain NaNs.
        In that case the original array is returned.
    """
    if masked_array is not None and masked_array.dtype.kind == 'f':
        check_class(masked_array, ma.masked_array)
        logger.debug("Replacing fill_values by NaNs")
        masked_array[:] = ma.filled(masked_array, np.nan)
        masked_array.set_fill_value(np.nan)
    else:
        return masked_array


#TODO: does recordMask help here?
# https://docs.scipy.org/doc/numpy/reference/maskedarray.baseclass.html#numpy.ma.MaskedArray.recordmask
def maskedEqual(array, missingValue):
    """ Mask an array where equal to a given (missing)value.

        Unfortunately ma.masked_equal does not work with structured arrays. See:
        https://mail.scipy.org/pipermail/numpy-discussion/2011-July/057669.html

        If the data is a structured array the mask is applied for every field (i.e. forming a
        logical-and). Otherwise, ma.masked_equal is called.
    """
    if array_is_structured(array):
        # Enforce the array to be masked
        if not isinstance(array, ma.MaskedArray):
            array = ma.MaskedArray(array)

        if missingValue is not None:
            # Set the mask separately per field. Do this only when a missing value is specified.
            # Otherwise, it might fail if the missing value is not of the correct type.
            for nr, field in enumerate(array.dtype.names):

                if hasattr(missingValue, '__len__'):
                    fieldMissingValue = missingValue[nr]
                else:
                    fieldMissingValue = missingValue

                try:
                    array[field] = ma.masked_equal(array[field], fieldMissingValue)
                except ValueError as ex:
                    # Type mismatch between missing value and field dtype
                    logger.warning("Missing values can't be determined for field: {}. Reason: {}"
                                   .format(field, ex))

        check_class(array, ma.MaskedArray) # post-condition check
        return array
    else:
        # masked_equal works with missing is None
        result = ma.masked_equal(array, missingValue, copy=False)
        check_class(result, ma.MaskedArray) # post-condition check
        return result



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository TreeItem (RTI) classes
    Tree items for use in the RepositoryTreeModel
"""
import logging
import os
import re

from argos.external import six
from argos.info import DEBUGGING
from argos.qt.treeitems import AbstractLazyLoadTreeItem
from argos.repo.iconfactory import RtiIconFactory
from argos.utils.cls import check_class, is_a_sequence, is_a_color_str
from argos.utils.dirs import normRealPath
from argos.utils.defs import DIM_TEMPLATE

logger = logging.getLogger(__name__)


def lengthToSummary(length, postfix="elements"):
    """ Constructs a summary from the length of an array
    """
    return str(length) + " " + postfix


def shapeToSummary(shape, postfix="elements"):
    """ Constructs a summary from the shape of an array
    """
    return " × ".join([str(elem) for elem in shape]) + " " + postfix


class BaseRti(AbstractLazyLoadTreeItem):
    """ TreeItem for use in a RepositoryTreeModel. (RTI = Repository TreeItem)
        Base node from which to derive the other types of nodes.

        Serves as an interface but can also be instantiated for debugging purposes.
    """
    _defaultIconGlyph = None  # Can be overridden by defining a _iconGlyph attribute
    _defaultIconColor = None  # Can be overridden by defining a _iconColor attribute

    def __init__(self, nodeName, iconColor, fileName=''):
        """ Constructor

            :param nodeName: name of this node (used to construct the node path).
            :param fileName: absolute path to the file where the data of this RTI originates.
        """
        super(BaseRti, self).__init__(nodeName=nodeName)
        assert is_a_color_str(iconColor), \
            "Icon color for {!r} not a hex string: {!r}".format(self, iconColor)
        self._iconColor = iconColor

        self._isOpen = False
        self._exception = None # Any exception that may occur when opening this item.

        check_class(fileName, six.string_types, allow_none=True)
        if fileName:
            fileName = os.path.abspath(fileName)
        self._fileName = fileName


    @classmethod
    def createFromFileName(cls, fileName, iconColor):
        """ Creates a BaseRti (or descendant), given a file name.
        """
        logger.debug("createFromFileName {}, {}, color={}".format(cls, fileName, iconColor))
        # See https://julien.danjou.info/blog/2013/guide-python-static-class-abstract-methods
        #logger.debug("Trying to create object of class: {!r}".format(cls))
        basename = os.path.basename(normRealPath(fileName)) # strips trailing slashes
        if not basename:
            logger.warning("Empty file name in path: {}. Using '<root>' as root path.")
            basename = '<root directory>'
        return cls(nodeName=basename, fileName=fileName, iconColor=iconColor)


    @property
    def fileName(self):
        """ Returns the name of the underlying the file.
        """
        return self._fileName


    def finalize(self):
        """ Can be used to cleanup resources. Should be called explicitly.
            Recursively calls the close method on all children and then on itself.
            In turn, close calls _closeRecources; descendants should override the latter.
        """
        for child in self.childItems:
            child.finalize()
        self.close()


    @property
    def isOpen(self):
        "Returns True if the underlying resources are opened"
        return self._isOpen


    def open(self):
        """ Opens underlying resources and sets isOpen flag.
            It calls _openResources. Descendants should usually override the latter
            function instead of this one.
        """
        self.clearException()
        try:
            if self._isOpen:
                logger.warning("Resources already open. Closing them first before opening.")
                self._closeResources()
                self._isOpen = False

            assert not self._isOpen, "Sanity check failed: _isOpen should be false"
            logger.debug("Opening {}".format(self))
            self._openResources()
            self._isOpen = True

            if self.model:
                self.model.sigItemChanged.emit(self)
            else:
                logger.warning("Model not set yet: {}".format(self))

        except Exception as ex:
            if DEBUGGING:
                raise
            logger.exception("Error during tree item open: {}".format(ex))
            self.setException(ex)


    def _openResources(self):
        """ Can be overridden to open the underlying resources.
            The default implementation does nothing.
            Is called by self.open
        """
        pass


    def close(self):
        """ Closes underlying resources and un-sets the isOpen flag.
            Any exception that occurs is caught and put in the exception property.
            This method calls _closeResources, which does the actual resource cleanup. Descendants
            should typically override the latter instead of this one.
        """
        self.clearException()
        try:
            if self._isOpen:
                logger.debug("Closing {}".format(self))
                self._closeResources()
                self._isOpen = False
            else:
                logger.debug("Resources already closed (ignored): {}".format(self))

            if self.model:
                self.model.sigItemChanged.emit(self)
            else:
                logger.warning("Model not set yet: {}".format(self))

        except Exception as ex:
            if DEBUGGING:
                raise
            logger.error("Error during tree item close: {}".format(ex))
            self.setException(ex)


    def _closeResources(self):
        """ Can be overridden to close the underlying resources.
            The default implementation does nothing.
            Is called by self.close
        """
        pass


    def _checkFileExists(self):
        """ Verifies that the underlying file exists and sets the _exception attribute if not
            Returns True if the file exists.
            If self._fileName is None, nothing is checked and True is returned.
        """
        if self._fileName and not os.path.exists(self._fileName):
            msg = "File not found: {}".format(self._fileName)
            logger.error(msg)
            self.setException(IOError(msg))
            return False
        else:
            return True

    @property
    def exception(self):
        """ The exception if an error has occurred during reading
        """
        return self._exception


    def setException(self, ex):
        """ Sets the exception attribute.
        """
        self._exception = ex


    def clearException(self):
        """ Forgets any stored exception to clear the possible error icon
        """
        self._exception = None


    def fetchChildren(self):
        """ Creates child items and returns them.
            Opens the tree item first if it's not yet open.

            Descendants should override _fetchAllChildren, not fetchChildren.
        """
        assert self._iconColor is not None, "Icon color none for: {}".format(self)
        assert self._canFetchChildren, "canFetchChildren must be True"
        try:
            self.clearException()

            if not self.isOpen:
                self.open() # Will set self._exception in case of failure

            if not self.isOpen:
                logger.warning("Opening item failed during fetch (aborted)")
                return [] # no need to continue if opening failed.

            childItems = []
            try:
                childItems = self._fetchAllChildren()
                assert is_a_sequence(childItems), "ChildItems must be a sequence"

            except Exception as ex:
                # This can happen, for example, when a NCDF/HDF5 file contains data types that
                # are not supported by the Python library that is used to read them.
                if DEBUGGING:
                    raise
                logger.error("Unable fetch tree item children: {}".format(ex))
                self.setException(ex)

            return childItems
        finally:
            self._canFetchChildren = False


    def _fetchAllChildren(self):
        """ The function that actually fetches the children. Default returns no children.
        """
        return []


    @property
    def iconColor(self):
        """ Returns the color of the icon (.e.g. '#FF0000' for red).
            :rtype: string
        """
        return self._iconColor


    @property
    def iconGlyph(self):
        """ Returns the kind of the icon (e.g. RtiIconFactory.FILE, RtiIconFactory.ARRAY, etc).
            The base implementation returns the default glyph of the class.
            :rtype: string
        """
        if hasattr(self, "_iconGlyph"):
            return getattr(self, "_iconGlyph")
        else:
            return self._defaultIconGlyph


    @property
    def decoration(self):
        """ The displayed icon.

            Shows open icon when node was visited (children are fetched). This allows users
            for instance to collapse a directory node but still see that it was visited, which
            may be useful if there is a huge list of directories.
        """
        rtiIconFactory = RtiIconFactory.singleton()

        if self._exception:
            return rtiIconFactory.getIcon(rtiIconFactory.ERROR, isOpen=False,
                                          color=rtiIconFactory.COLOR_ERROR)
        else:
            return rtiIconFactory.getIcon(self.iconGlyph, isOpen=not self.canFetchChildren(),
                                          color=self._iconColor)

    @property
    def isSliceable(self):
        """ Returns True if the underlying data can be sliced.
            An inspector should always check this before using an index/slice on an RTI.

            The base implementation returns False. Descendants should override this if they contain
            an array that can be sliced.
        """
        return False


    def __getitem__(self, index):
        """ Called when using the RTI with an index (e.g. rti[0]).

            The base function is abstract. Descendants should override this if they contain
            an array that can be sliced (i.e. self.isSliceable is True). It should then
            call __getitem__(index) on the underlying array data.

            This may return a numpy array or scalars. Collector.getSlicedArray will convert this into an array
            if case a scalar is expected.
        """
        raise NotImplemented("Override for slicable arrays")


    @property
    def nDims(self): # TODO: rename to numDims?
        """ The number of dimensions of the underlying array
            The base implementation returns len(self.arrayShape). Descendants may override this to
            provide a more efficient implementation
        """
        return len(self.arrayShape)


    @property
    def arrayShape(self):
        """ Returns the shape of the underlying array.
            The base function is abstract. Descendants should override this if they contain
            an array that can be sliced (i.e. self.isSliceable is True).
        """
        raise NotImplemented("Override for slicable arrays")


    @property
    def chunking(self):
        """ List with chunk sizes if chunked storage is used. Or 'contiguous' for contiguous storage

            By default the empty string is returned, meaning chunking is not applicable for this RTI
        """
        return ""


    @property
    def chunksString(self):
        """ String representation of the chunk sizes. For convenience.
        """
        chunking = self.chunking
        if isinstance(chunking, str):
            return chunking
        else:
            return " × ".join(str(elem) for elem in chunking)


    @property
    def dimensionality(self):
        """ String that describes if the RTI is an array, scalar, field, etc.
        """
        return ""


    @property
    def elementTypeName(self):
        """ String representation of the element type.
        """
        return ""


    @property
    def typeName(self):
        """ String representation of the type. By default, the elementTypeName + dimensionality.
        """
        return self.elementTypeName + " " + self.dimensionality


    @property
    def attributes(self):
        """ The attribute dictionary.
            The attributes generally contain meta data about the item.
        """
        return {}


    @property
    def dimensionNames(self):
        """ Returns a list with the name of each of the RTI's dimensions.
            The default implementation returns ['Dim0', 'Dim1', ...] by default. Descendants can
            override this.
        """
        return [DIM_TEMPLATE.format(dimNr) for dimNr in range(self.nDims)] # TODO: cache?


    @property
    def dimensionPaths(self):
        """ Returns a list with the full path names of the dimensions.
            The default implementation returns None. Descendants can override this.
        """
        return None


    @property
    def unit(self):
        """ Returns the unit of the RTI. The base implementation returns ''.
        """
        return ""


    @property
    def summary(self):
        """ Returns a summary of the contents of the RTI.  E.g. 'array 20 x 30' elements.

            The base implementation returns ''.
        """
        return ""


    def quickLook(self, width: int):
        """ Returns a string representation fof the RTI to use in the Quik Look pane.

            Args:
                Width in characters.
        """
        if not self.isSliceable:
            return ""
        else:
            # We would like to use np.array2string so that we can use commas to separate the array
            # elements, and can set the line length without side effects, but unfortunately
            # np.array2string doesn't work for masked arrays.
            # See: https://github.com/numpy/numpy/issues/11674
            return str(self[...])


    @property
    def missingDataValue(self):
        """ Returns the value to indicate missing data.

            The base implementation returns None, indicating that there is no missing data
            specified.
        """
        return None




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Base class for inspectors
"""
import logging

from argos.config.abstractcti import ResetMode
from argos.config.groupcti import MainGroupCti
from argos.info import DEBUGGING
from argos.qt import QtWidgets, QtSignal
from argos.utils.cls import type_name, check_class
from argos.widgets.constants import DOCK_SPACING, DOCK_MARGIN
from argos.widgets.display import MessageDisplay
from argos.widgets.misc import BasePanel

logger = logging.getLogger(__name__)

class InvalidDataError(Exception):
    """ Exception that should be raised if the inspector cannot handle this type of data.
        Can be used to distuingish the situation from other exceptions an then, for example,
        draw an empty plot instead of the error message pane.
    """
    pass


class UpdateReason(object):
    """ Enumeration class that contains the reason for updating the inspector contents.
    """
    NEW_MAIN_WINDOW     = "new main window"
    INSPECTOR_CHANGED   = "inspector changed"
    RTI_CHANGED         = "repo tree item changed"
    COLLECTOR_COMBO_BOX = "collector combobox changed"
    COLLECTOR_SPIN_BOX  = "collector spinbox changed"
    CONFIG_CHANGED      = "config changed"

    __VALID_REASONS = (NEW_MAIN_WINDOW, INSPECTOR_CHANGED, RTI_CHANGED,
                       COLLECTOR_COMBO_BOX, COLLECTOR_SPIN_BOX, CONFIG_CHANGED)


    @classmethod
    def validReasons(cls):
        """ Returns a tuple with all valid reasond
        """
        return cls.__VALID_REASONS


    @classmethod
    def checkValid(cls, reason):
        """ Raises ValueError if the reason is not one of the valid enumerations
        """
        if reason not in cls.__VALID_REASONS:
            raise ValueError("reason must be one of {}, got {}".format(cls.__VALID_REASONS, reason))



class AbstractInspector(QtWidgets.QStackedWidget):
    """ Abstract base class for inspectors.
        An inspector is a stacked widget; it has a contents page and and error page.
    """
    _fullName = "base" # see the fullName() class method for explanation
    ERROR_PAGE_IDX = 0
    CONTENTS_PAGE_IDX = 1

    sigShowMessage = QtSignal(str)
    sigUpdated = QtSignal(bool)  # Parameter indicates sucess

    def __init__(self, collector, parent=None):
        """ Constructor.

            When subclassing the AbstractInspector try to minimize the work done in the constructor.
            Place all functionality that may raise an exception in updateContents or clearContents,
            these functions will show an error page and prevent the application from aborting. If
            an exception occurs in the constructor it is not caught!

            :param collector: the data collector from where this inspector gets its data
            :param parent: parent widget.
        """

        super(AbstractInspector, self).__init__(parent)

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self._config = MainGroupCti(nodeName='inspector') # Is typically redefined.
        self._collector = collector

        self.errorWidget = MessageDisplay()
        self.addWidget(self.errorWidget)

        self.contentsWidget = BasePanel()
        self.addWidget(self.contentsWidget)

        self.contentsLayout = QtWidgets.QBoxLayout(QtWidgets.QBoxLayout.TopToBottom)
        self.contentsLayout.setSpacing(DOCK_SPACING)
        self.contentsLayout.setContentsMargins(DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN)
        self.contentsWidget.setLayout(self.contentsLayout)

        self.setCurrentIndex(self.CONTENTS_PAGE_IDX)


    def finalize(self):
        """ Is called before destruction. Can be used to clean-up resources
        """
        logger.debug("Finalizing: {}".format(self))


    @property
    def config(self):
        """ The root config tree item for this inspector.
        """
        return self._config


    @property
    def windowNumber(self):
        """ The instance number of the window this inspector belongs to.
        """
        return self.collector.windowNumber


    @classmethod
    def descriptionHtml(cls):
        """ A long description that will be displayed as help in the inspector-open dialog box.
        """
        return ""


    @classmethod
    def axesNames(cls):
        """ The names of the axes that this inspector visualizes.

            This determines the dimensionality of the inspector. For example an inspector that shows
            an image, and has axes names 'X' and 'Y', is 2-dimensional. Also the slicedArray that
            will be returned by the collector will then be 2-dimensional. This invariant also holds
            when the axesNames is an empty tuple. The slicedArray will then be zero-dimensional.
            For a 0-D example, see qtplugins.text.TextInspector.

            The axesNames should be row-first, so in this example: ['Y', 'X']

            The names should not include the string "Axis"; the Collector.fullAxisNames()
            returns that.
        """
        return tuple()


    @property
    def collector(self):
        """ The data collector from where this inspector gets its data
        """
        return self._collector


    def configValue(self, nodePath):
        """ Returns the config value data at the node path
        """
        return self.config.findByNodePath(nodePath).configValue


    def updateContents(self, reason=None, initiator=None):
        """ Tries to draw the widget contents with the updated RTI.
            Shows the error page in case an exception is raised while drawing the contents.
            Descendants should override _drawContents, not updateContents.

            During the call of _drawContents, the updating of the configuration tree is blocked to
            avoid circular effects. After that, a call to self.config.refreshFromTarget() is
            made to refresh the configuration tree with possible new values from the inspector
            (the inspector is the configuration's target, hence the name).

            The reason parameter is a string (one of the UpdateReason values) that indicates why
            the inspector contents was updated. This can, for instance, be used to optimize
            drawing the inspector contents. Note that the reason may be undefined (None).

            The initiator may contain the object that initiated the updated. The type depends on the
            reason. At the moment the initiator is only implemented for the "config changed" reason. In
            this case the initiator will be the Config Tree Item (CTI that has changed).
        """
        UpdateReason.checkValid(reason)
        logger.debug("---- Inspector updateContents, reason: {}, initiator: {}"
                     .format(reason, initiator))
        logger.debug("Inspector: {}".format(self))
        logger.debug("RTI: {}".format(self.collector.rti))
        try:
            self.setCurrentIndex(self.CONTENTS_PAGE_IDX)

            wasBlocked = self.config.model.setRefreshBlocked(True)
            try:
                self._drawContents(reason=reason, initiator=initiator)
                logger.debug("_drawContents finished successfully")

                # Update the config tree from the (possibly) new state of the PgLinePlot1d inspector,
                # e.g. the axis range may have changed while drawing.
                # self.config.updateTarget() # TODO: perhaps enable this here (instead of doing it in the inspector._drawContents when needed)?
            finally:
                self.config.model.setRefreshBlocked(wasBlocked)

            # Call refreshFromTarget in case the newly applied configuration resulted in a change
            # of the state of the configuration's target's (i.e. the inspector state)
            logger.debug("_drawContents finished successfully, calling refreshFromTarget...")
            self.config.refreshFromTarget()
            logger.debug("refreshFromTarget finished successfully")

            self.sigUpdated.emit(True)

        except InvalidDataError as ex:
            if str(ex):
                # Exception message can be empty in common cases. Don't show the message label then.
                logger.debug("Unable to draw the inspector contents: '{}'".format(ex))
                self.sigShowMessage.emit(str(ex))

            self.sigUpdated.emit(True)

        except Exception as ex:####
            if DEBUGGING:
                raise

            logger.error("Error while drawing the inspector: {} ----".format(ex))
            logger.exception(ex)
            self._clearContents()
            self.setCurrentIndex(self.ERROR_PAGE_IDX)
            self._showError(msg=str(ex), title=type_name(ex))

            self.sigUpdated.emit(False)

        else:
            logger.debug("---- updateContents finished successfully")


    def _resetRequired(self, reason, _initiator=None):
        """ Uses reason parameter (and optionally the _initiator) to determine axis must be reset.
        """
        return (self.config.model.autoReset and
                (reason == UpdateReason.RTI_CHANGED or reason == UpdateReason.COLLECTOR_COMBO_BOX))


    def resetConfig(self):
        """ Resets config depending on the config reset mode.
        """
        resetMode = self.config.model.resetMode
        if resetMode == ResetMode.All:
            self.config.resetToDefault()
        elif resetMode == ResetMode.Ranges:
            self.config.resetRangesToDefault()
        else:
            raise ValueError("Unexpected config mode")



    def _drawContents(self, reason=None, initiator=None):
        """ Is called by updateContents to do the actual drawing.
            Descendants should override _drawContents and not worry about exceptions;
            the updateContents will show the error page if an exception is raised.

            See the updateContents() docstring for the reason and initiator parameters.
        """
        raise NotImplementedError()


    def _clearContents(self):
        """ Override to clear the widget contents.

            Is called to empty the inspector when an exception occurs during _drawContents,
            to prevent further errors.
        """
        raise NotImplementedError()


    def _showError(self, msg="", title="Error"):
        """ Shows an error message.
        """
        self.errorWidget.setError(msg=msg, title=title)




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Stores for representing data that is read from text files.
"""
from __future__ import absolute_import

import logging, os
import numpy as np

from numpy.lib.npyio import NpzFile

from argos.qt import QtWidgets
from argos.repo.iconfactory import RtiIconFactory, ICON_COLOR_UNDEF
from argos.repo.memoryrtis import ArrayRti, SliceRti, MappingRti
from argos.utils.cls import check_is_an_array, check_class

logger = logging.getLogger(__name__)


# Do not allow pickle in numpy.load(), at least for now. This can be a security risk
ALLOW_PICKLE = False


class NumpyTextFileRti(ArrayRti):
    """ Reads a 2D array from a simple text file using numpy.loadtxt().
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(NumpyTextFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                               iconColor=iconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._array = np.loadtxt(self._fileName, ndmin=0)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None
        self.attributes.clear()


    def _fetchAllChildren(self):
        """ Adds an ArrayRti per column as children so that they can be inspected easily
        """
        childItems = []
        _nRows, nCols = self._array.shape if self._array is not None else (0, 0)
        for col in range(nCols):
            colItem = SliceRti(self._array[:, col], nodeName="column-{}".format(col),
                               fileName=self.fileName, iconColor=self.iconColor,
                               attributes=self.attributes)
            childItems.append(colItem)
        return childItems



class NumpyBinaryFileRti(ArrayRti):
    """ Reads a single Numpy array from a binary file (.npy) using numpy.load().

        The file must have been saved with numpy.save() A TypeError is raised if this is not the

        The allow_pickle is set to False, therefore no object arrays can be read.
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(NumpyBinaryFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                                 iconColor=iconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children

            Returns True so that the file can be opened, even though the array has no children.
        """
        return True


    def _openResources(self):
        """ Uses numpy.load to open the underlying file
        """
        arr = np.load(self._fileName, allow_pickle=ALLOW_PICKLE)
        check_is_an_array(arr)
        self._array = arr


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None



class NumpyCompressedFileRti(MappingRti):
    """ Reads one ore more arrays from a Numpy zip file (.npz) using numpy.load().

        The file must have been saved with numpy.savez() A TypeError is raised if this is not the
        case.

        The allow_pickle is set to False, therefore no object arrays can be read.
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(NumpyCompressedFileRti, self).__init__(None,
                                                     nodeName=nodeName, fileName=fileName,
                                                     iconColor=iconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.load to open the underlying file
        """
        dct = np.load(self._fileName, allow_pickle=ALLOW_PICKLE)
        check_class(dct, NpzFile)
        self._dictionary = dct


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._dictionary = None




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the RtiIconFactory
"""
from __future__ import print_function
import logging, os

from argos.qt import Qt, QtCore, QtGui, QtWidgets, QtSvg
from argos.info import program_directory, DEBUGGING
from argos.utils.cls import check_class

logger = logging.getLogger(__name__)

ICON_COLOR_UNDEF      = '#FFFFFF'  # White to indicate something went wrong
ICON_COLOR_ERROR      = '#FF0000'
ICON_COLOR_UNKNOWN    = '#999999'
#ICON_COLOR_MEMORY     = '#FFAA00'
#ICON_COLOR_MEMORY     = '#FFFF40'
ICON_COLOR_MEMORY     = '#FFDD20'


class RtiIconFactory(object):
    """ A factory class that generates QIcons for use in the Repository Tree.

        Typically the RepoTreeItems have a reference to a (singleton) RtiIconFactory and
        call getIcon to get an icon of a certain glyph, color and open/close state.

        The factory contains a icon registry that contains for registered (glyph, open/close state)
        combinations the location of an SVG file. The SVG files originate from the snip-icon
        library. See http://www.snipicons.com/

        The getIcon method can optionally set the fill color of the icons. It also caches the
        generated QItem objects to that they only have to be created once.
    """

    ICONS_DIRECTORY = os.path.join(program_directory(), 'img/snipicons')
    #ICON_SIZE = 32 # Render in this size

    # File state
    OPEN = "open"
    CLOSED = "closed"

    # Registered glyph names
    TRANSPARENT = "transparent"  # Transparent icon to reserve space
    ERROR = "error"
    FOLDER = "folder"
    FILE = "file"
    ARRAY = "array"
    FIELD = "field"
    DIMENSION = "dimension"
    SEQUENCE = "sequence"
    SCALAR = "scalar"

    # Icon colors from constants defined at the top of this module
    COLOR_UNDEF      = ICON_COLOR_UNDEF
    COLOR_UNKNOWN    = ICON_COLOR_UNKNOWN
    COLOR_ERROR      = ICON_COLOR_ERROR
    COLOR_MEMORY     = ICON_COLOR_MEMORY

    _singleInstance = None

    def __init__(self):
        """ Constructor
        """
        self._icons = {}
        self._registry = {}
        self.colorsToBeReplaced = ('#008BFF', '#00AAFF')
        self.renderSizes = [16, 24, 32, 64]

        self.registerIcon(None, None) # no icon
        self.registerIcon("",   None) # no icon
        self.registerIcon("transparent_1x1.svg", self.TRANSPARENT)
        self.registerIcon("warning-sign.svg", self.ERROR)
        self.registerIcon("folder-open.svg",  self.FOLDER, True)
        self.registerIcon("folder-close.svg", self.FOLDER, False)
        self.registerIcon("file.svg",         self.FILE, True)
        self.registerIcon("file-inverse.svg", self.FILE, False)
        self.registerIcon("th-large.svg",     self.ARRAY)
        self.registerIcon("asterisk.svg",     self.FIELD)
        self.registerIcon("move.svg",         self.DIMENSION)
        self.registerIcon("align-left.svg",   self.SEQUENCE)
        self.registerIcon("leaf.svg",         self.SCALAR)


    @classmethod
    def singleton(cls):
        """ Returns the RtiIconFactory singleton.
        """
        if cls._singleInstance is None:
            cls._singleInstance = cls()
        return cls._singleInstance


    def registerIcon(self, fileName, glyph, isOpen=None):
        """ Register an icon SVG file given a glyph, and optionally the open/close state.

            :param fileName: filename to the SVG file.
                If the filename is a relative path, the ICONS_DIRECTORY will be prepended.
            :param glyph: a string describing the glyph (e.g. 'file', 'array')
            :param isOpen: boolean that indicates if the RTI is open or closed.
                If None, the icon will be registered for open is both True and False
            :return: QIcon
        """
        check_class(isOpen, bool, allow_none=True)

        if fileName and not os.path.isabs(fileName):
            fileName = os.path.join(self.ICONS_DIRECTORY, fileName)

        if isOpen is None:
            # Register both opened and closed variants
            self._registry[(glyph, True)] = fileName
            self._registry[(glyph, False)] = fileName
        else:
            self._registry[(glyph, isOpen)] = fileName


    def getIcon(self, glyph, isOpen, color=None):
        """ Returns a QIcon given a glyph name, open/closed state and color.

            The reslulting icon is cached so that it only needs to be rendered once.

            :param glyph: name of a registered glyph (e.g. 'file', 'array')
            :param isOpen: boolean that indicates if the RTI is open or closed.
            :param color: '#RRGGBB' string (e.g. '#FF0000' for red)
            :return: QtGui.QIcon
        """
        try:
            fileName = self._registry[(glyph, isOpen)]
        except KeyError:
            logger.warning("Unregistered icon glyph: {} (open={})".format(glyph, isOpen))
            from argos.utils.logs import log_dictionary
            log_dictionary(self._registry, "registry", logger=logger)
            raise

        return self.loadIcon(fileName, color=color)


    def loadIcon(self, fileName, color=None):
        """ Reads SVG from a file name and creates an QIcon from it.

            Optionally replaces the color. Caches the created icons.

            :param fileName: absolute path to an icon file.
                If False/empty/None, None returned, which yields no icon.
            :param color: '#RRGGBB' string (e.g. '#FF0000' for red)
            :return: QtGui.QIcon
        """
        if not fileName:
            return None

        key = (fileName, color)
        if key not in self._icons:
            try:
                with open(fileName, 'r') as input:
                    svg = input.read()

                self._icons[key] = self.createIconFromSvg(svg, color=color)
            except Exception as ex:
                # It's preferable to show no icon in case of an error rather than letting
                # the application fail. Icons are a (very) nice to have.
                logger.warning("Unable to read icon: {}".format(ex))
                if DEBUGGING:
                    raise
                else:
                    return None

        return self._icons[key]


    def createIconFromSvg(self, svg, color=None, colorsToBeReplaced=None):
        """ Creates a QIcon given an SVG string.

            Optionally replaces the colors in colorsToBeReplaced by color.

            :param svg: string containing Scalable Vector Graphics XML
            :param color: '#RRGGBB' string (e.g. '#FF0000' for red)
            :param colorsToBeReplaced: optional list of colors to be replaced by color
                If None, it will be set to the fill colors of the snip-icon libary
            :return: QtGui.QIcon
        """
        if colorsToBeReplaced is None:
            colorsToBeReplaced = self.colorsToBeReplaced

        if color:
            for oldColor in colorsToBeReplaced:
                svg = svg.replace(oldColor, color)

        # From http://stackoverflow.com/questions/15123544/change-the-color-of-an-svg-in-qt
        qByteArray = QtCore.QByteArray(svg.encode('utf-8'))
        # qByteArray = QtCore.QByteArray() # the old PyQt way
        # qByteArray.append(svg)

        svgRenderer = QtSvg.QSvgRenderer(qByteArray)
        icon = QtGui.QIcon()
        for size in self.renderSizes:
            pixMap = QtGui.QPixmap(QtCore.QSize(size, size))
            pixMap.fill(Qt.transparent)
            pixPainter = QtGui.QPainter(pixMap)
            pixPainter.setRenderHint(QtGui.QPainter.TextAntialiasing, True)
            pixPainter.setRenderHint(QtGui.QPainter.Antialiasing, True)
            svgRenderer.render(pixPainter)
            pixPainter.end()
            icon.addPixmap(pixMap)

        return icon



    def createIconFromSvg(self, svg, color=None, colorsToBeReplaced=None):
        """ Creates a QIcon given an SVG string.

            Optionally replaces the colors in colorsToBeReplaced by color.

            :param svg: string containing Scalable Vector Graphics XML
            :param color: '#RRGGBB' string (e.g. '#FF0000' for red)
            :param colorsToBeReplaced: optional list of colors to be replaced by color
                If None, it will be set to the fill colors of the snip-icon libary
            :return: QtGui.QIcon
        """
        if colorsToBeReplaced is None:
            colorsToBeReplaced = self.colorsToBeReplaced

        if color:
            for oldColor in colorsToBeReplaced:
                svg = svg.replace(oldColor, color)

        # From http://stackoverflow.com/questions/15123544/change-the-color-of-an-svg-in-qt
        qByteArray = QtCore.QByteArray(svg.encode('utf-8'))
        # qByteArray = QtCore.QByteArray() # the old PyQt way
        # qByteArray.append(svg)

        svgRenderer = QtSvg.QSvgRenderer(qByteArray)
        icon = QtGui.QIcon()
        for size in self.renderSizes:
            pixMap = QtGui.QPixmap(QtCore.QSize(size, size))
            pixMap.fill(Qt.transparent)
            pixPainter = QtGui.QPainter(pixMap)
            pixPainter.setRenderHint(QtGui.QPainter.TextAntialiasing, True)
            pixPainter.setRenderHint(QtGui.QPainter.Antialiasing, True)
            svgRenderer.render(pixPainter)
            pixPainter.end()
            icon.addPixmap(pixMap)

        return icon



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository tree items that are read using import routines of Scipy
    See http://docs.scipy.org/doc/scipy-0.16.0/reference/io.html
"""
from __future__ import absolute_import

import logging, os
import scipy
import scipy.io
import scipy.io.wavfile

from argos.repo.memoryrtis import ArrayRti, SliceRti, MappingRti
from argos.repo.iconfactory import RtiIconFactory, ICON_COLOR_UNDEF

logger = logging.getLogger(__name__)


class MatlabFileRti(MappingRti):
    """ Read data from a MATLAB file using the scipy.io.loadmat function.

        Note: v4 (Level 1.0), v6 and v7 to 7.2 matfiles are supported.

        From version 7.3 onward matfiles are stored in HDF-5 format. You can read them with the
        Argos hDF-5 plugin (which uses on h5py)
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(MatlabFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                            iconColor=iconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._dictionary = scipy.io.loadmat(self._fileName)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._dictionary = None



class IdlSaveFileRti(MappingRti):
    """ ReadS data from an IDL 'save file'.

        Uses scipy.io.readsav to read the file. This reads all data at once when the file
        is open (in contrast to lazy loading each node separately). Therefore, it may take a while
        to read a large save-file. During this time the application is not responsive!
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an MappingRti with None as underlying dictionary.
        """
        super(IdlSaveFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                             iconColor=iconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file
        """
        self._dictionary = scipy.io.readsav(self._fileName)


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._dictionary = None



class WavFileRti(ArrayRti):
    """ Read data from a WAV file using the scipy.io.wavfile.read function

        First tries to read with memory mapping, if that fails it tries to read the file without
        memory mapping.
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', fileName='', iconColor=ICON_COLOR_UNDEF):
        """ Constructor. Initializes as an ArrayRTI with None as underlying array.
        """
        super(WavFileRti, self).__init__(None, nodeName=nodeName, fileName=fileName,
                                         iconColor=iconColor)
        self._checkFileExists()


    def hasChildren(self):
        """ Returns True if the item has (fetched or unfetched) children
        """
        return True


    def _openResources(self):
        """ Uses numpy.loadtxt to open the underlying file.
        """
        try:
            rate, data = scipy.io.wavfile.read(self._fileName, mmap=True)
        except Exception as ex:
            logger.warning(ex)
            logger.warning("Unable to read wav with memmory mapping. Trying without now.")
            rate, data = scipy.io.wavfile.read(self._fileName, mmap=False)

        self._array = data
        self.attributes['rate'] = rate


    def _closeResources(self):
        """ Closes the underlying resources
        """
        self._array = None
        self.attributes.clear()


    def _fetchAllChildren(self):
        """ Adds an ArrayRti per column as children so that they can be inspected easily
        """
        childItems = []
        if self._array.ndim == 2:
            _nRows, nCols = self._array.shape if self._array is not None else (0, 0)
            for col in range(nCols):
                colItem = SliceRti(self._array[:, col], nodeName="channel-{}".format(col),
                                   fileName=self.fileName, iconColor=self.iconColor,
                                   attributes=self.attributes)
                childItems.append(colItem)
        return childItems




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Miscellaneous functions and classes using Qt
"""
from __future__ import print_function

import logging
import os.path

from argos.qt import QtWidgets

logger = logging.getLogger(__name__)


def setWidgetSizePolicy(widget, hor=None, ver=None):
    """ Sets horizontal and/or vertical size policy on a widget
    """
    sizePolicy = widget.sizePolicy()
    logger.debug("widget {} size policy Befor: {} {}"
                 .format(widget, sizePolicy.horizontalPolicy(), sizePolicy.verticalPolicy()))

    if hor is not None:
        sizePolicy.setHorizontalPolicy(hor)

    if ver is not None:
        sizePolicy.setVerticalPolicy(ver)

    widget.setSizePolicy(sizePolicy)

    sizePolicy = widget.sizePolicy()
    logger.debug("widget {} size policy AFTER: {} {}"
                 .format(widget, sizePolicy.horizontalPolicy(), sizePolicy.verticalPolicy()))


def processEvents():
    """ Processes all pending events for the calling thread until there are no more events to
        process.
    """
    QtWidgets.QApplication.instance().processEvents()


def setApplicationQtStyle(styleName):
    """ Sets the Qt style (e.g. to 'fusion')
    """
    qApp = QtWidgets.QApplication.instance()
    logger.debug("Setting Qt style to: {}".format(styleName))
    qApp.setStyle(QtWidgets.QStyleFactory.create(styleName))
    if qApp.style().objectName().lower() != styleName.lower():
        logger.warning(
            "Setting style failed: actual style {!r} is not the specified style {!r}"
            .format(qApp.style().objectName(), styleName))


def setApplicationStyleSheet(fileName):
    """ Reads the style sheet from file and set it as application style sheet.
    """
    fileName = os.path.abspath(fileName)
    logger.debug("Reading qss from: {}".format(fileName))
    try:
        with open(fileName) as input:
            qss = input.read()
    except Exception as ex:
        logger.warning("Unable to read style sheet from '{}'. Reason: {}".format(fileName, ex))
        return

    QtWidgets.QApplication.instance().setStyleSheet(qss)



class BasePanel(QtWidgets.QFrame):
    """ Base panel from which others are derived.

        Define shape
    """
    def __init__(self, **kwargs):
        super(BasePanel, self).__init__(**kwargs)
        #self.setFrameShape(QtWidgets.QFrame.StyledPanel)
        #self.setFrameShadow(QtWidgets.QFrame.Sunken)





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

r""" Default config and log directories under the different platforms.

    Config files are stored in a subdirectory of the genericConfigLocation. Log files are stored
    in a sub directory of the genericLocalDataLocation. These are:

        Windows:
            baseConfigLocation    -> C:\Users\<user>\AppData\Local\
            baseLocalDataLocation -> C:\Users\<user>\AppData\Local\

        OS-X:
            baseConfigLocation    -> ~/Library/Preferences
            baseLocalDataLocation -> ~/Library/Application Support

        Linux:
            baseConfigLocation    -> ~/.config
            baseLocalDataLocation -> ~/.local/share

    See http://doc.qt.io/qt-5/qsettings.html#platform-specific-notes.

"""

import logging
import os.path
import platform

from argos.info import ORGANIZATION_NAME, SCRIPT_NAME

logger  = logging.getLogger(__name__)


def normRealPath(path):
    """ Returns the normalized real path.

        If the path is empty or None it is returned as-is. This is to prevent expanding to the
        current directory in case of undefined paths.
    """
    if path:
        return os.path.normpath(os.path.realpath(path))
    else:
        return path


def ensureDirectoryExists(dirName):
    """ Creates a directory if it doesn't yet exist.
    """
    if not os.path.exists(dirName):
        logger.info("Creating directory: {}".format(normRealPath(dirName)))
        os.makedirs(dirName)


def ensureFileExists(pathName):
    """ Creates an empty file file if it doesn't yet exist. Also creates necessary directory path.

        :returns: the normRealPath of the path name.
    """
    pathName = normRealPath(pathName)
    dirName, fileName = os.path.split(pathName)
    ensureDirectoryExists(dirName)

    if not os.path.exists(pathName):
        logger.info("Creating empty file: {}".format(pathName))
        with open(pathName, 'w') as f:
            f.write('')

    # Check file exists and is not a directory.
    assert os.path.isfile(pathName), "File does not exist or is a directory: {!r}".format(pathName)
    return pathName


def homeDirectory():
    """ Returns the user's home directory.

        https://stackoverflow.com/a/4028943/625350
    """
    return os.path.expanduser("~")





################
# Config files #
################


def baseConfigLocation():
    r""" Gets the base configuration directory (for all applications of the user).

        See the module doc string at the top for details.
    """
    # Same as QtCore.QStandardPaths.AppConfigLocation, but without having to import Qt
    sysName = platform.system()

    if sysName == "Darwin":
        configDir = os.path.join(homeDirectory(), 'Library', 'Preferences')
    elif sysName == "Linux":
        configDir = os.path.join(homeDirectory(), '.config')
    elif sysName == "Windows":
        configDir = os.environ.get("LOCALAPPDATA", os.path.join(homeDirectory(), 'AppData', 'Local'))
    else:
        raise AssertionError("Unknown Operating System: {}".format(sysName))

    assert configDir, "No baseConfigLocation found."
    return normRealPath(configDir)


def argosConfigDirectory():
    r""" Gets the Argos configuration directory.

        The config directory is platform dependent. (See the module doc string at the top).
    """
    return os.path.join(baseConfigLocation(), ORGANIZATION_NAME, SCRIPT_NAME)

#
# def appConfigFileName(configFile):
#     """ Returns default config file if configFile is emtpy string or None
#     """
#     if configFile:
#         # Config file specified on the command line
#         return configFile
#     else:
#         # Use the default config file name
#         configDir = argosConfigDirectory()
#         configFile = normRealPath(os.path.join(configDir, 'config.yaml'))
#         return configFile



#############
# Log files #
#############


def baseLocalDataLocation():
    r""" Gets the base configuration directory (for all applications of the user).

        The config directory is platform dependent (see the Qt documentation for baseDataLocation).
        On Windows this will be something like:
            C:\Users\<user>\AppData\Local\

        See the module doc string at the top for details.
    """
    # Same as QtCore.QStandardPaths.AppConfigLocation, but without having to import Qt
    sysName = platform.system()

    if sysName == "Darwin":
        cfgDir = os.path.join(homeDirectory(), 'Library', 'Application Support')
    elif sysName == "Linux":
        cfgDir = os.path.join(homeDirectory(), '.local', 'share')
    elif sysName == "Windows":
        cfgDir = os.environ.get("APPLOCALDATA2", os.path.join(homeDirectory(), 'AppData', 'Local'))
    else:
        raise AssertionError("Unknown Operating System: {}".format(sysName))

    assert cfgDir, "No baseConfigLocation found."
    return normRealPath(cfgDir)


def argosLocalDataDirectory():
    r""" Returns a directory where Argos can store files locally (not roaming).

        This directory is platform dependent. (See the module doc string at the top).
    """
    return os.path.join(baseLocalDataLocation(), ORGANIZATION_NAME, SCRIPT_NAME)



def argosLogDirectory():
    r""" Returns the directory where Argos can store its log files.

        This is the 'logs' subdirectory of the argosLocalDataDirectory()
    """
    return os.path.join(argosLocalDataDirectory(), 'logs')






# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository tree.
"""
from __future__ import print_function

import logging
from argos.qt import QtWidgets, QtGui, QtCore, QtSignal, QtSlot, Qt
from argos.config.groupcti import MainGroupCti
from argos.config.boolcti import BoolCti
from argos.repo.detailplugins.attr import AttributesPane
from argos.repo.detailplugins.prop import PropertiesPane
from argos.repo.detailplugins.quicklook import QuickLookPane
from argos.repo.registry import globalRtiRegistry
from argos.repo.repotreemodel import RepoTreeModel
from argos.widgets.argostreeview import ArgosTreeView
from argos.widgets.constants import LEFT_DOCK_WIDTH, DOCK_SPACING, DOCK_MARGIN, COL_KIND_WIDTH
from argos.widgets.constants import COL_NODE_NAME_WIDTH, COL_ELEM_TYPE_WIDTH, COL_SUMMARY_WIDTH

from argos.widgets.misc import BasePanel


logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901



class RepoWidget(BasePanel):
    """ Groups the repository tree plus the details dock widgets.
    """
    def __init__(self, repoTreeModel, collector, parent=None):
        """ Constructor.
            :param parent:
        """
        super(RepoWidget, self).__init__(parent=parent)

        self.detailDockPanes = []

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(DOCK_SPACING)
        self.mainLayout.setContentsMargins(DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN)
        self.setLayout(self.mainLayout)

        self.mainSplitter = QtWidgets.QSplitter(orientation=Qt.Vertical)
        self.mainLayout.addWidget(self.mainSplitter)

        self.repoTreeView = RepoTreeView(repoTreeModel, collector)
        self.mainSplitter.addWidget(self.repoTreeView)
        self.mainSplitter.setCollapsible(0, False)

        self.tabWidget = QtWidgets.QTabWidget()
        self.mainSplitter.addWidget(self.tabWidget)
        self.mainSplitter.setCollapsible(1, True)

        self.propertiesPane = self.addDetailsPane(PropertiesPane(self.repoTreeView))
        self.attributesPane = self.addDetailsPane(AttributesPane(self.repoTreeView))
        self.quickLookPane = self.addDetailsPane(QuickLookPane(self.repoTreeView))

        self.repoTreeView.sigRepoItemChanged.connect(self.repoItemChanged)
        self.tabWidget.currentChanged.connect(self.tabChanged)

        self.tabWidget.setCurrentIndex(2) # Show quick look tab the first time the program runs



    def addDetailsPane(self, detailPane):
        """ Adds a pane widget as a tab.

            Returns: detailPane so calls can be chained.
        """
        self.tabWidget.addTab(detailPane, detailPane.classLabel())
        self.detailDockPanes.append(detailPane)
        return detailPane


    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        cfg = dict(
            tabIndex = self.tabWidget.currentIndex(),
            splitterSizes = self.mainSplitter.sizes(),
            treeHeaders = self.repoTreeView.marshall(),
            propertiesPane = self.propertiesPane.marshall(),
            attributesPane = self.attributesPane.marshall(),
        )

        return cfg


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        if 'tabIndex' in cfg:
            self.tabWidget.setCurrentIndex(cfg['tabIndex'])

        if 'splitterSizes' in cfg:
            self.mainSplitter.setSizes(cfg['splitterSizes'])

        self.repoTreeView.unmarshall(cfg.get('treeHeaders'))

        self.propertiesPane.unmarshall(cfg.get('propertiesPane', {}))
        self.attributesPane.unmarshall(cfg.get('attributesPane', {}))


    def repoItemChanged(self, rti):
        """ Updates the current tab with the newly selected repoWidget

            The rti parameter can be None when no RTI is selected in the repository tree.
        """
        logger.debug("RepoWidget.repoItemChanged: {}".format(rti))

        curPanel = self.tabWidget.currentWidget() # can be None according to docs
        if curPanel:
            curPanel.repoItemChanged(rti)
        else:
            logger.debug("No panel selected")


    def tabChanged(self, _tabIndex):
        """ Updates the tab from the currently selected repo item in the tree.
        """
        currentRepoItem, _currentIndex = self.repoTreeView.getCurrentItem()
        self.repoItemChanged(currentRepoItem)



class RepoTreeView(ArgosTreeView):
    """ Tree widget for browsing the data repository.

        Currently it supports only selecting one item. That is, the current item is always the
        selected item (see notes in ArgosTreeView documentation for details).
    """
    # sigRepoItemChanged parameter is BaseRti or None when no RTI is selected, or the model is empty
    sigRepoItemChanged = QtSignal(object)

    def __init__(self, repoTreeModel, collector, parent=None):
        """ Constructor.

            Maintains a reference to a collector. The repo tree view updates the collector when
            the currentIndex changes.
        """
        super(RepoTreeView, self).__init__(treeModel=repoTreeModel, parent=parent)

        self._collector = collector
        self._config = self._createConfig()

        treeHeader = self.header()
        treeHeader.resizeSection(RepoTreeModel.COL_NODE_NAME, COL_NODE_NAME_WIDTH)
        treeHeader.resizeSection(RepoTreeModel.COL_KIND, COL_KIND_WIDTH)
        treeHeader.resizeSection(RepoTreeModel.COL_ELEM_TYPE, COL_ELEM_TYPE_WIDTH)
        treeHeader.resizeSection(RepoTreeModel.COL_SUMMARY, COL_SUMMARY_WIDTH)

        # Note: this setting is stored in the json file. Remove the config file if you want to
        # change it.
        treeHeader.setStretchLastSection(True)

        headerNames = self.model().horizontalHeaders
        enabled = dict((name, True) for name in headerNames)
        enabled[headerNames[RepoTreeModel.COL_NODE_NAME]] = False # Cannot be unchecked
        checked = dict((name, False) for name in headerNames)
        checked[headerNames[RepoTreeModel.COL_NODE_NAME]] = True
        checked[headerNames[RepoTreeModel.COL_KIND]] = True
        checked[headerNames[RepoTreeModel.COL_ELEM_TYPE]] = True
        checked[headerNames[RepoTreeModel.COL_SUMMARY]] = True
        self.addHeaderContextMenu(checked=checked, enabled=enabled, checkable={})

        self.setContextMenuPolicy(Qt.DefaultContextMenu) # will call contextMenuEvent
        self.setUniformRowHeights(True)
        #self.setIconSize(QtCore.QSize(16, 16))

        # Add actions
        self.currentItemActionGroup = QtWidgets.QActionGroup(self)
        self.currentItemActionGroup.setExclusive(False)

        removeFileAction = QtWidgets.QAction("Remove from Tree", self.currentItemActionGroup,
                                         shortcut=QtGui.QKeySequence.Delete,
                                         triggered=self.removeCurrentItem)
        self.addAction(removeFileAction)

        reloadFileAction = QtWidgets.QAction("Reload File", self.currentItemActionGroup,
                                         shortcut=QtGui.QKeySequence.Refresh,   #"Ctrl+R",
                                         triggered=self.reloadFileOfCurrentItem)
        self.addAction(reloadFileAction)

        self.openItemAction = QtWidgets.QAction("Open Item", self, triggered=self.openCurrentItem)
        self.addAction(self.openItemAction)

        self.closeItemAction = QtWidgets.QAction("Close Item", self, triggered=self.closeCurrentItem)
        self.addAction(self.closeItemAction)

        self.collapseItemAction = QtWidgets.QAction("Collapse Item", self, triggered=self.collapseCurrentItem)
        self.addAction(self.collapseItemAction)

        self.copyPathAction = QtWidgets.QAction("Copy Path to Clipboard", self,
                                                triggered=self.copyPathToClipboard)
        self.addAction(self.copyPathAction)

        # Connect signals
        selectionModel = self.selectionModel() # need to store reference to prevent crash in PySide
        selectionModel.currentChanged.connect(self.currentItemChanged)

        # Close files on collapse. Note that, self.collapsed does NOT seem to be connected to self.collapse by default,
        # so there is not conflict here. Also there is no need to connect to expand, this is automatic with the
        # fetchMore mechanism
        self.collapsed.connect(self.closeItem)

        self.model().sigItemChanged.connect(self.repoTreeItemChanged)
        self.model().sigAllChildrenRemovedAtIndex.connect(self.collapse)


    def finalize(self):
        """ Disconnects signals and frees resources
        """
        self.model().sigItemChanged.disconnect(self.repoTreeItemChanged)

        selectionModel = self.selectionModel() # need to store reference to prevent crash in PySide
        selectionModel.currentChanged.disconnect(self.currentItemChanged)


    def contextMenuEvent(self, event):
        """ Creates and executes the context menu for the tree view
        """
        menu = QtWidgets.QMenu(self)

        for action in self.actions():
            menu.addAction(action)

        openAsMenu = QtWidgets.QMenu(parent=menu)
        openAsMenu.setTitle("Open Item As")
        openAsMenu.aboutToShow.connect(lambda :self._populateOpenAsMenu(openAsMenu))
        menu.insertMenu(self.closeItemAction, openAsMenu) # Insert before "Close Item" entry.

        menu.exec_(event.globalPos())


    def _populateOpenAsMenu(self, openAsMenu):
        """ Repopulates the submenu for the Open Item choice (which is used to reload files).
        """
        registry = globalRtiRegistry()
        for rtiRegItem in (registry.items + registry.extraItemsForOpenAsMenu()):

            if not rtiRegItem.triedImport:
                rtiRegItem.tryImportClass()

            def createTrigger():
                """Function to create a closure with the regItem"""
                _rtiRegItem = rtiRegItem # keep reference in closure
                return lambda: self.reloadFileOfCurrentItem(_rtiRegItem)

            action = QtWidgets.QAction("{}".format(rtiRegItem.name), self,
                enabled=bool(rtiRegItem.successfullyImported is not False),
                triggered=createTrigger(), icon=rtiRegItem.decoration)
            openAsMenu.addAction(action)

        return openAsMenu


    @property
    def config(self):
        """ The root config tree item for this widget
        """
        return self._config


    def _createConfig(self):
        """ Creates a config tree item (CTI) hierarchy containing default children.
        """
        # Currently not added to the config tree as there are no configurable items.
        rootItem = MainGroupCti('data repository')
        rootItem.insertChild(BoolCti('test checkbox', False)) # Does nothing yet
        return rootItem


    @property
    def collector(self): # TODO: move to selector class in the future
        """ The collector that this selector view will update. Read only property.
        """
        return self._collector


    def sizeHint(self):
        """ The recommended size for the widget."""
        return QtCore.QSize(LEFT_DOCK_WIDTH, 450)


    @QtSlot()
    def copyPathToClipboard(self):
        """ Copies the path of the currently selected item to the clipboard
        """
        currentItem, currentIndex = self.getCurrentItem()
        if not currentIndex.isValid():
            return

        QtWidgets.QApplication.clipboard().setText(currentItem.nodePath)
        logger.info("Copied path to the clipboard: {}".format(currentItem.nodePath))



    @QtSlot()
    def openCurrentItem(self):
        """ Opens the current item in the repository.
        """
        logger.debug("openCurrentItem called")
        _currentItem, currentIndex = self.getCurrentItem()
        if not currentIndex.isValid():
            return

        # Expanding the node will indirectly call RepoTreeModel.fetchMore which will call
        # BaseRti.fetchChildren, which will call BaseRti.open and thus open the current RTI.
        # BaseRti.open will emit the self.model.sigItemChanged signal, which is connected to
        # RepoTreeView.repoTreeItemChanged.
        self.expand(currentIndex)


    @QtSlot()
    def closeCurrentItem(self):
        """ Closes the current item in the repository.
            All its children will be unfetched and closed.
        """
        self.closeItem(self.getRowCurrentIndex())


    @QtSlot(QtCore.QModelIndex)
    def closeItem(self, index):
        """ Closes the item at the index and collapses the node
        """
        logger.debug("closeItem called")
        if not index.isValid():
            logger.debug("Index invalid (returning)")
            return

        # First we remove all the children, this will close them as well.
        # It will emit sigAllChildrenRemovedAtIndex, which is connected to the collapse method of
        # all trees. It will thus collapse the current item in all trees. This is necessary,
        # otherwise the children will be fetched immediately.
        self.model().removeAllChildrenAtIndex(index)

        # Close the item. BaseRti.close will emit the self.model.sigItemChanged signal,
        # which is connected to RepoTreeView.repoTreeItemChanged.
        item = self.model().getItem(index)
        logger.debug("Item: {}".format(item))
        item.close()


    def expand(self, index):
        """ Expands current item. Updates context menu action.
        """
        super(RepoTreeView, self).expand(index)
        self.collapseItemAction.setEnabled(self.isExpanded(index))


    def collapse(self, index):
        """ Collapses current item. Updates context menu action.
        """
        super(RepoTreeView, self).collapse(index)
        self.collapseItemAction.setEnabled(self.isExpanded(index))


    @QtSlot()
    def collapseCurrentItem(self):
        """ Closes the current item in the repository.
            All its children will be unfetched and closed.
        """
        currentIndex = self.getRowCurrentIndex()
        oldBlockState = self.blockSignals(True)  # Prevent automatically closing of item.
        try:
            self.collapse(currentIndex)
        finally:
            self.blockSignals(oldBlockState)


    @QtSlot()
    def removeCurrentItem(self):
        """ Removes the current item from the repository tree.
        """
        logger.debug("removeCurrentFile")
        currentIndex = self.getRowCurrentIndex()
        if not currentIndex.isValid():
            return

        self.model().deleteItemAtIndex(currentIndex) # this will close the items resources.


    @QtSlot()
    def reloadFileOfCurrentItem(self, rtiRegItem=None):
        """ Finds the repo tree item that holds the file of the current item and reloads it.
            Reloading is done by removing the repo tree item and inserting a new one.

            The new item will have by of type rtiRegItem.cls. If rtiRegItem is None (the default),
            the new rtiClass will be the same as the old one.
            The rtiRegItem.cls will be imported. If this fails the old class will be used, and a
            warning will be logged.
        """
        logger.debug("reloadFileOfCurrentItem, rtiClass={}".format(rtiRegItem))

        currentIndex = self.getRowCurrentIndex()
        if not currentIndex.isValid():
            return

        currentItem, _ = self.getCurrentItem()
        oldPath = currentItem.nodePath

        fileRtiIndex = self.model().findFileRtiIndex(currentIndex)
        isExpanded = self.isExpanded(fileRtiIndex)

        newRtiIndex = self.model().reloadFileAtIndex(fileRtiIndex, rtiRegItem=rtiRegItem)

        try:
            # Expand and select the name with the old path
            _lastItem, lastIndex = self.expandPath(oldPath)
            self.setCurrentIndex(lastIndex)
            return lastIndex
        except Exception as ex:
            # The old path may not exist anymore. In that case select file RTI
            logger.warning("Unable to select {!r} beause of: {}".format(oldPath, ex))
            self.setExpanded(newRtiIndex, isExpanded)
            self.setCurrentIndex(newRtiIndex)
            return newRtiIndex


    @QtSlot(QtCore.QModelIndex)
    @QtSlot(QtCore.QModelIndex, QtCore.QModelIndex)
    def currentItemChanged(self, currentIndex, _previousIndex=None):
        """ Enables/disables actions when a new item is the current item in the tree view.

            Is not called currentChanged as this would override an existing method. We want to
            connect this to the currentChanged signal at the end of the constructor, which would
            then not be possible.
        """
        self.currentRepoTreeItemChanged()


    def repoTreeItemChanged(self, rti):
        """ Called when repo tree item has changed (the item itself, not a new selection)

            If the item is the currently selected item, the collector (inspector) and
            metadata widgets are updated.
        """
        logger.debug("repoTreeItemChanged: {}".format(rti))
        currentItem, _currentIndex = self.getCurrentItem()

        if rti == currentItem:
            self.currentRepoTreeItemChanged()
        else:
            logger.debug("Ignoring changed item as is not the current item: {}".format(rti))


    def currentRepoTreeItemChanged(self):
        """ Called to update the GUI when a repo tree item has changed or a new one was selected.
        """
        # When the model is empty the current index may be invalid and the currentItem may be None.
        currentItem, currentIndex = self.getCurrentItem()

        hasCurrent = currentIndex.isValid()
        assert hasCurrent == (currentItem is not None), \
            "If current index is valid, currentIndex may not be None" # sanity check

        # Set the item in the collector, will subsequently update the inspector.
        if hasCurrent:
            logger.debug("Adding rti to collector: {}".format(currentItem.nodePath))
            self.collector.setRti(currentItem)
            #if rti.asArray is not None: # TODO: maybe later, first test how robust it is now
            #    self.collector.setRti(rti)

        # Update context menus in the repo tree
        self.currentItemActionGroup.setEnabled(hasCurrent)
        isTopLevel = hasCurrent and self.model().isTopLevelIndex(currentIndex)
        self.collapseItemAction.setEnabled(self.isExpanded(currentIndex))
        self.openItemAction.setEnabled(currentItem is not None
                                       and currentItem.hasChildren()
                                       and not currentItem.isOpen)
        self.closeItemAction.setEnabled(currentItem is not None
                                        and currentItem.hasChildren()
                                        and currentItem.isOpen)

        # Emit sigRepoItemChanged signal so that, for example, details panes can update.
        logger.debug("Emitting sigRepoItemChanged: {}".format(currentItem))
        self.sigRepoItemChanged.emit(currentItem)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" RTI attributes inspector.
"""
import logging

from argos.qt import QtWidgets
from argos.repo.detailpanes import DetailTablePane
from argos.utils.cls import to_string, type_name
from argos.widgets.constants import COL_SUMMARY_WIDTH

logger = logging.getLogger(__name__)



class AttributesPane(DetailTablePane):
    """ Shows the attributes of the selected repo tree item
    """
    _label = "Attributes"

    HEADERS = ["Name", "Value", "Type"]
    (COL_ATTR_NAME, COL_VALUE, COL_ELEM_TYPE) = range(len(HEADERS))

    def __init__(self, repoTreeView, parent=None):
        super(AttributesPane, self).__init__(repoTreeView, parent=parent)
        self.table.setWordWrap(True)
        self.table.addHeaderContextMenu(enabled = {'Name': False, 'Value': False},
                                        checked = {'Type': False})

        tableHeader = self.table.horizontalHeader()
        tableHeader.resizeSection(self.COL_ATTR_NAME, 125)
        tableHeader.resizeSection(self.COL_VALUE, 150)
        tableHeader.resizeSection(self.COL_ELEM_TYPE, COL_SUMMARY_WIDTH)


    def _drawContents(self, currentRti=None):
        """ Draws the attributes of the currentRTI
        """
        #logger.debug("_drawContents: {}".format(currentRti))
        table = self.table
        table.setUpdatesEnabled(False)
        try:
            table.clearContents()
            verticalHeader = table.verticalHeader()
            verticalHeader.setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

            attributes = currentRti.attributes if currentRti is not None else {}
            table.setRowCount(len(attributes))

            for row, (attrName, attrValue) in enumerate(sorted(attributes.items())):
                attrStr = to_string(attrValue, decode_bytes='utf-8')

                try:
                    type_str = type_name(attrValue)
                except Exception as ex:
                    logger.exception(ex)
                    type_str = "<???>"

                nameItem = QtWidgets.QTableWidgetItem(attrName)
                nameItem.setToolTip(attrName)
                table.setItem(row, self.COL_ATTR_NAME, nameItem)
                valItem = QtWidgets.QTableWidgetItem(attrStr)
                valItem.setToolTip(attrStr)
                table.setItem(row, self.COL_VALUE, valItem)
                table.setItem(row, self.COL_ELEM_TYPE, QtWidgets.QTableWidgetItem(type_str))
                table.resizeRowToContents(row)

            verticalHeader.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        finally:
            table.setUpdatesEnabled(True)







""" 3rd party modules and packages included in argos.
"""
import logging
import os

logger = logging.getLogger(__name__)
#
# # Set the ARGOS_CMLIB_PATH to the path of the CmLib if you want to use another version of this
# # package, for instance during development. E.g export ARGOS_CMLIB_PATH=/Users/kenter/prog/py/cmlib/
# cmLibPath = os.environ.get("ARGOS_CMLIB_PATH")
#
# if cmLibPath is not None:
#     logger.debug("Prepending to external package path: {}".format(cmLibPath))
#     __path__.insert(0, cmLibPath)





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Text widget inspector.
"""
import logging

from argos.config.groupcti import MainGroupCti, GroupCti
from argos.config.choicecti import ChoiceCti
from argos.config.qtctis import FontCti
from argos.inspector.abstract import AbstractInspector
from argos.qt import QtGui, QtWidgets
from argos.utils.cls import to_string, check_class
from argos.widgets.constants import MONO_FONT, FONT_SIZE

logger = logging.getLogger(__name__)


class TextInspectorCti(MainGroupCti):
    """ Configuration tree item for a TextInspectorCti inspector
    """
    def __init__(self, textInspector, nodeName):
        """ Constructor

            :param textInspector: the TextInspector widget that is being configured
            :param nodeName: node name
        """
        super(TextInspectorCti, self).__init__(nodeName)

        check_class(textInspector, TextInspector)
        self.textInspector = textInspector

        Opt = QtGui.QTextOption
        self.wordWrapCti = self.insertChild(
            ChoiceCti('word wrap',
                      displayValues=['No wrapping', 'Word boundaries', 'Anywhere',
                                     'Boundaries or anywhere'],
                      configValues=[Opt.NoWrap, Opt.WordWrap, Opt.WrapAnywhere,
                                    Opt.WrapAtWordBoundaryOrAnywhere]))

        self.encodingCti = self.insertChild(
            ChoiceCti('encoding', editable=True,
                      configValues=['utf-8', 'ascii', 'latin-1', 'windows-1252']))

        self.fontCti = self.insertChild(FontCti(self.textInspector.editor, "font",
                                                defaultData=QtGui.QFont(MONO_FONT, FONT_SIZE)))


class TextInspector(AbstractInspector):
    """ Displays the data of a single array element in a large text-widget.

        This is useful when the data under inspection is a large, multi-line text string.
    """
    def __init__(self, collector, parent=None):

        super(TextInspector, self).__init__(collector, parent=parent)

        self.editor = QtWidgets.QPlainTextEdit()
        self.editor.setReadOnly(True)

        self.contentsLayout.addWidget(self.editor)

        self._config = self._createConfig()


    @classmethod
    def axesNames(cls):
        """ The names of the axes that this inspector visualizes.
            See the parent class documentation for a more detailed explanation.

            Returns an empty tuple; the data under inspection is zero-dimensional.
        """
        return tuple()


    def _createConfig(self):
        """ Creates a config tree item (CTI) hierarchy containing default children.
        """
        return TextInspectorCti(textInspector=self, nodeName='text')


    def _clearContents(self):
        """ Clears the inspector widget when no valid input is available.
        """
        self.editor.clear()


    def _drawContents(self, reason=None, initiator=None):
        """ Converts the (zero-dimensional) sliced array to string and puts it in the text editor.

            The reason and initiator parameters are ignored.
            See AbstractInspector.updateContents for their description.
        """
        logger.debug("TextInspector._drawContents: {}".format(self))

        # If auto-reset is true, reset config complete or partially, depending on the mode.
        if self._resetRequired(reason, initiator):
            self.resetConfig()

        self._clearContents()

        slicedArray = self.collector.getSlicedArray()

        if slicedArray is None:
            return

        # Sanity check, the slicedArray should be zero-dimensional. It can be used as a scalar.
        # In fact, using an index (e.g. slicedArray[0]) will raise an exception.
        assert slicedArray.data.ndim == 0, \
            "Expected zero-dimensional array. Got: {}".format(slicedArray.ndim)

        # Valid data from here...
        maskedArr = slicedArray.asMaskedArray() # So that we call mask[()] for boolean masks
        slicedScalar = maskedArr[()] # Convert to Numpy scalar
        isMasked = maskedArr.mask[()]

        text = to_string(slicedScalar, masked=isMasked, maskFormat='--',
                         decode_bytes=self.config.encodingCti.configValue)
        self.editor.setPlainText(text)
        self.editor.setWordWrapMode(self.config.wordWrapCti.configValue)

        # Update the editor font from the font config item (will call self.editor.setFont)
        self.config.updateTarget()



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Argos Color Library
"""
from __future__ import print_function

import logging
import os.path

from os import listdir

from cmlib import CmLib, CmLibModel, DATA_DIR

from argos.utils.cls import SingletonMixin

logger = logging.getLogger(__name__)

# The color maps that are favorites then the program is started for the first time or reset.
DEF_FAV_COLOR_MAPS = [
    'SciColMaps/Oleron', 'SciColMaps/Nuuk', 'SciColMaps/Acton', 'CET/CET-CBL2', 'MatPlotLib/Gray',
    'CET/CET-C2', 'CET/CET-R2', 'MatPlotLib/BrBG', 'MatPlotLib/Tab20', 'MatPlotLib/Magma',
    'MatPlotLib/Tab10', 'MatPlotLib/Cubehelix', 'MatPlotLib/Viridis', 'MatPlotLib/Coolwarm']

DEFAULT_COLOR_MAP = "MatPlotLib/Magma"
assert DEFAULT_COLOR_MAP in DEF_FAV_COLOR_MAPS, "Default color map not in default favorites."


class CmLibSingleton(CmLib, SingletonMixin):

    def __init__(self, **kwargs):
        super(CmLibSingleton, self).__init__(**kwargs)

        logger.debug("CmLib singleton: {}".format(self))

        cmDataDir = DATA_DIR
        logger.info("Importing color map library from: {}".format(cmDataDir))

        # Don't import from Color Brewer since those are already included in MatPlotLib.
        # With sub-sampling the color maps similar maps can be achieved as the Color Brewer maps.
        excludeList = ['ColorBrewer2']
        for path in listdir(cmDataDir):
            if path in excludeList:
                logger.debug("Not importing catalogue from exlude list: {}".format(excludeList))
                continue

            fullPath = os.path.join(cmDataDir, path)
            if os.path.isdir(fullPath):
                self.load_catalog(fullPath)

        logger.debug("Number of color maps: {}".format(len(self.color_maps)))


class CmLibModelSingleton(CmLibModel, SingletonMixin):

    def __init__(self, **kwargs):
        super(CmLibModelSingleton, self).__init__(CmLibSingleton.instance(), **kwargs)




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the ConfigItemDelegate class.
"""
from __future__ import print_function

import logging
from argos.utils.defs import InvalidInputError
from argos.qt import Qt, QtCore, QtWidgets
from argos.qt.misc import widgetSubCheckBoxRect
from argos.widgets.constants import TREE_CELL_SIZE_HINT

logger = logging.getLogger(__name__)

# Qt classes have many ancestors
#pylint: disable=R0901

class ConfigItemDelegate(QtWidgets.QStyledItemDelegate):
    """ Provides editing facilities for config tree items.
        Creates an editor kd on the underlying config tree item at an index.

        We don't use a QItemEditorFactory since that is typically registered for a type of
        QVariant. We then would have to make a new UserType QVariant for (each?) CTI.
        This is cumbersome and possibly unPyQTtonic :-)
    """
    def __init__(self, parent=None):
        super(ConfigItemDelegate, self).__init__(parent=parent)

        self.commitData.connect(self._prepareCommit)


    def createEditor(self, parent, option, index):
        """ Returns the widget used to change data from the model and can be reimplemented to
            customize editing behavior.

            Reimplemented from QStyledItemDelegate.
        """
        logger.debug("ConfigItemDelegate.createEditor, parent: {!r}".format(parent.objectName()))
        assert index.isValid(), "sanity check failed: invalid index"

        cti = index.model().getItem(index)
        editor = cti.createEditor(self, parent, option)
        return editor


    def finalizeEditor(self, editor):
        """ Calls editor.finalize().

            Not part of the QAbstractItemView interface but added to be able to free resources.

            Note that, unlike the other methods of this class, finalizeEditor does not have an
            index parameter. We cannot derive this since indexForEditor is a private method in Qt.
            Therefore a AbstractCtiEditor maintains a reference to its config tree item (cti).
        """
        editor.finalize()


    def _prepareCommit(self, editor):
        """ Called when commitData signal is emitted. Calls the prepareCommit of the editor.
        """
        editor.prepareCommit()


    def setEditorData(self, editor, index):
        """ Provides the widget with data to manipulate.
            Calls the setEditorValue of the config tree item at the index.

            :type editor: QWidget
            :type index: QModelIndex

            Reimplemented from QStyledItemDelegate.
        """
        # We take the config value via the model to be consistent with setModelData
        data = index.model().data(index, Qt.EditRole)
        editor.setData(data)


    def setModelData(self, editor, model, index):
        """ Gets data from the editor widget and stores it in the specified model at the item index.
            Does this by calling getEditorValue of the config tree item at the index.

            :type editor: QWidget
            :type model: ConfigTreeModel
            :type index: QModelIndex

            Reimplemented from QStyledItemDelegate.
        """
        try:
            data = editor.getData()
        except InvalidInputError as ex:
            logger.warning(ex)
        else:
            # The value is set via the model so that signals are emitted
            logger.debug("ConfigItemDelegate.setModelData: {}".format(data))
            model.setData(index, data, Qt.EditRole)


    def updateEditorGeometry(self, editor, option, index):
        """ Ensures that the editor is displayed correctly with respect to the item view.
        """
        cti = index.model().getItem(index)
        if cti.checkState is None:
            displayRect = option.rect
        else:
            checkBoxRect = widgetSubCheckBoxRect(editor, option)
            offset = checkBoxRect.x() + checkBoxRect.width()
            displayRect = option.rect
            displayRect.adjust(offset, 0, 0, 0)

        editor.setGeometry(displayRect)





# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains the UntypedCti and UntypedCtiEditor classes
"""
import logging

from argos.config.abstractcti import AbstractCti, AbstractCtiEditor
from argos.qt import QtWidgets, Qt

logger = logging.getLogger(__name__)



class UntypedCti(AbstractCti):
    """ Config Tree Item to store a any type of data as long as it can be edited with a QLineEdit.

        Typically it's better to use 'typed' CTIs, where the data is always internally stored in
        the same type (enforced by _enforceDataType).

        This item is non-editable and can, for instance, be used for introspection.
    """
    def __init__(self, nodeName, defaultData='', doc=''):
        """ Constructor. For the parameters see the AbstractCti constructor documentation.
        """
        super(UntypedCti, self).__init__(nodeName, defaultData)
        self.doc = doc

    def _enforceDataType(self, value):
        """ Since UntypedCti can store any type of data no conversion will be done.
        """
        return value

    def createEditor(self, delegate, parent, option):
        """ Creates an UntypedCtiEditor.
            For the parameters see the AbstractCti constructor documentation.
            Note: since the item is not editable this will never be called.
        """
        return UntypedCtiEditor(self, delegate, parent=parent)

    @property
    def valueColumnItemFlags(self):
        """ Returns the flags determine how the user can interact with the value column.
            Returns Qt.NoItemFlags so that the item is not editable
        """
        return Qt.NoItemFlags


class UntypedCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a QLineEdit for editing UntypedCti objects.

        Only
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(UntypedCtiEditor, self).__init__(cti, delegate, parent=parent)
        self.lineEditor = self.addSubEditor(QtWidgets.QLineEdit(), isFocusProxy=True)

    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        self.lineEditor.setText(str(data))

    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.lineEditor.text()


